import os

from medcat.vocab import Vocab
from medcat.storage.serialisers import get_serialiser, deserialise

import numpy as np

import unittest
import tempfile

from . import UNPACKED_EXAMPLE_MODEL_PACK_PATH, RESOURCES_PATH
from . import UNPACKED_V1_MODEL_PACK_PATH


ZIPPED_VOCAB_PATH = os.path.join(RESOURCES_PATH, "mct2_vocab.zip")


class VocabCreationTests(unittest.TestCase):
    word1 = {
        "word": "WORD1", "cnt": 1,
    }
    word2 = {
        "word": "WORD2", "cnt": 1,
        "vec": np.array([0,   100, 100]),
    }
    word3 = {
        "word": "WORD3", "cnt": 1,
        "vec": np.array([100,   0, 100]),
        "replace": True,
    }
    word4 = {
        "word": "WORD4", "cnt": 1,
        "vec": np.array([100, 100,   0]),
        "replace": False,
    }
    word5 = {
        "word": "WORD5", "cnt": 3,
        "vec": None,
        "replace": False
    }
    word6 = {
        "word": "WORD6", "cnt": 10,
        "vec": np.array([50,  50,  25]),
        "replace": False
    }
    all_words = [
        word1, word2, word3, word4,
        # for negative sampling stuff
        word5, word6
    ]

    def setUp(self):
        self._temp_file = tempfile.NamedTemporaryFile()
        self.temp_file = self._temp_file.name
        self.vocab = Vocab()

    def tearDown(self):
        self._temp_file.close()

    def test_does_not_have_words_by_default(self):
        for word in self.all_words:
            self.assertNotIn(word["word"], self.vocab)

    def test_remembers_words(self):
        for word in self.all_words:
            self.vocab.add_word(**word)
            self.assertIn(word["word"], self.vocab)

    def test_can_overwrite_word(self):
        word = self.word3.copy()  # allows to overwrite
        self.vocab.add_word(**word)
        word_cp = word.copy()
        word_cp["vec"] = 2 * word["vec"]
        self.vocab.add_word(**word_cp)
        got_vec = self.vocab.vec(word["word"])
        self.assertTrue(np.all(got_vec == word_cp["vec"]))


class VocabSaveTests(unittest.TestCase):
    serialiser = get_serialiser('dill')
    all_words = VocabCreationTests.all_words

    @classmethod
    def setUpClass(cls):
        cls._temp_dir = tempfile.TemporaryDirectory()
        cls.target_file = os.path.join(cls._temp_dir.name, "vocab.dat")
        cls.vocab = Vocab()
        for word in cls.all_words:
            cls.vocab.add_word(**word)

    @classmethod
    def tearDownClass(cls):
        cls._temp_dir.cleanup()

    def _serialise_default(self):
        self.serialiser.serialise(self.vocab, self.target_file)

    def test_can_serialise(self):
        self._serialise_default()
        self.assertTrue(os.path.exists(self.target_file))

    def assert_equal_dicts(self, d1: dict, d2: dict, path: str = '') -> None:
        keys1, keys2 = d1.keys(), d2.keys()
        if keys1 != keys2:
            raise AssertionError(
                f"Incompatible keys for vocabs: {keys1} vs {keys2}")
        for k, v1 in d1.items():
            v2 = d2[k]
            if isinstance(v2, dict):
                self.assert_equal_dicts(v1, v2,  f"{path}.{k}" if path else k)
                continue
            if isinstance(v1, np.ndarray) or isinstance(v2, np.ndarray):
                is_eq = np.all(v1 == v2)
            else:
                is_eq = v1 == v2
            if not is_eq:
                raise AssertionError(
                    f"Incompatible values for {k}: {v1} vs {v2}")

    def assert_vocab_equals(self, vocab1: Vocab, vocab2: Vocab):
        self.assert_equal_dicts(vocab1.__dict__, vocab2.__dict__)

    def test_can_deserialise(self):
        self._serialise_default()
        loaded = self.serialiser.deserialise(self.target_file)
        self.assertIsInstance(loaded, Vocab)
        self.assert_vocab_equals(self.vocab, loaded)

    def test_deserialised_has_all_words(self):
        self._serialise_default()
        vocab = self.serialiser.deserialise(self.target_file)
        for word in self.all_words:
            with self.subTest(word):
                self.assertIn(word["word"], vocab)


class VocabTests(unittest.TestCase):
    serialiser = get_serialiser('dill')
    all_words = VocabCreationTests.all_words

    @classmethod
    def setUpClass(cls):
        cls.vocab = Vocab()
        for word in cls.all_words:
            cls.vocab.add_word(**word)
        cls.vocab.init_cumsums()

    def test_neg_sampling_gets_list_of_ints(self, num_to_get: int = 10):
        neg_samples = self.vocab.get_negative_samples(num_to_get)
        self.assertIsInstance(neg_samples, list)
        self.assertEqual(len(neg_samples), num_to_get)
        for inum, index in enumerate(neg_samples):
            with self.subTest(f"INDEX: {index} @ {inum}"):
                self.assertIsInstance(index, int)

    def test_neg_sampling_does_not_include_vectorless(
            self, num_to_get: int = 30):
        inds = self.vocab.get_negative_samples(num_to_get)
        for index in inds:
            with self.subTest(f"Index: {index}"):
                # in the right list
                self.assertIn(index, self.vocab.vec_index2word)
                word = self.vocab.vec_index2word[index]
                info = self.vocab.vocab[word]
                # the info has vector
                self.assertIn("vector", info)
                # the vector is an array or a list
                self.assertIsInstance(self.vocab.vec(word), (np.ndarray, list))


class DefaultVocabTests(unittest.TestCase):
    VOCAB_PATH = os.path.join(UNPACKED_EXAMPLE_MODEL_PACK_PATH, 'vocab')
    LEGACY_VOCAB_PATH = os.path.join(UNPACKED_V1_MODEL_PACK_PATH, "vocab.dat")
    EXP_SHAPE = (7,)

    @classmethod
    def setUpClass(cls):
        cls.vocab: Vocab = deserialise(cls.VOCAB_PATH)

    # NOTE: the MCTv1 vocab has a vector (for 'chronic')
    #       that is longer than the rest (the reast are 7 length,
    #       this one is 8 length). So we want to make sure everything
    #       is in order here.
    def test_has_correct_vectors(self):
        for w, info in self.vocab.vocab.items():
            with self.subTest(w):
                self.assertEqual(info['vector'].shape, self.EXP_SHAPE)

    def test_convenience_save(self):
        with tempfile.TemporaryDirectory() as dir:
            self.vocab.save(dir)
            self.assertTrue(os.path.exists(dir))
            obj = deserialise(dir)
            self.assertIsInstance(obj, Vocab)

    def test_convenience_load(self):
        vocab = Vocab.load(self.VOCAB_PATH)
        self.assertIsInstance(vocab, Vocab)

    def test_can_load_from_zip(self):
        vocab = Vocab.load(ZIPPED_VOCAB_PATH)
        self.assertIsInstance(vocab, Vocab)

    def test_can_convert_legacy_upon_load(self):
        loaded = Vocab.load(self.LEGACY_VOCAB_PATH)
        self.assertIsInstance(loaded, Vocab)

    def test_can_save_to_zip(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, 'vocab.zip')
            # NOTE: auto detection will write as zip
            self.vocab.save(file_name)
            self.assertTrue(os.path.exists(file_name))
            self.assertTrue(os.path.isfile(file_name))
            # and can load from saved zip
            loaded = Vocab.load(file_name)
            self.assertIsInstance(loaded, Vocab)



import logging
import torch
from typing import Any, Optional, Union, cast
from torch import nn
from transformers import PretrainedConfig
from transformers.modeling_utils import PreTrainedModel

from medcat.config.config_rel_cat import ConfigRelCAT
from medcat.components.addons.relation_extraction.config import (
    RelExtrBaseConfig)
from medcat.components.addons.relation_extraction.ml_utils import (
    create_dense_layers, get_annotation_schema_tag)


logger = logging.getLogger(__name__)


class BaseModelBluePrint(nn.Module):
    """ Base class for the RelCAT models
    """

    hf_model: PreTrainedModel
    relcat_config: ConfigRelCAT
    model_config: PretrainedConfig
    drop_out: nn.Dropout
    fc1: nn.Linear
    fc2: nn.Linear
    fc3: nn.Linear

    def __init__(self, pretrained_model_name_or_path: str,
                 relcat_config: ConfigRelCAT,
                 model_config: Union[PretrainedConfig,
                                     RelExtrBaseConfig]):
        """ Class to hold the HF model + model_config

        Args:
            pretrained_model_name_or_path (str): path to load the model from,
                    this can be a HF model i.e: "bert-base-uncased",
                    if left empty, it is normally assumed that a model is
                    loaded from 'model.dat' using the RelCAT.load() method.
                    So if you are initializing/training a model from scratch
                    be sure to base it on some model.
            relcat_config (ConfigRelCAT): relcat config.
            model_config (PretrainedConfig): HF bert config for model.
        """
        super().__init__()

    def forward(
            self,
            input_ids: Optional[torch.Tensor] = None,
            attention_mask: Optional[torch.Tensor] = None,
            token_type_ids: Optional[torch.Tensor] = None,
            position_ids: Any = None,
            head_mask: Any = None,
            encoder_hidden_states: Any = None,
            encoder_attention_mask: Any = None,
            Q: Any = None,
            e1_e2_start: Any = None,
            pooled_output: Any = None
            ) -> Optional[tuple[torch.Tensor, torch.Tensor]]:
        """Forward pass for the model

        Args:
            input_ids (torch.Tensor): input token ids. Defaults to None.
            attention_mask (torch.Tensor): attention mask for the input ids.
                Defaults to None.
            token_type_ids (torch.Tensor): token type ids for the input ids.
                Defaults to None.
            position_ids (Any): The position IDs. Defaults to None.
            head_mask (Any): The head mask. Defaults to None.
            encoder_hidden_states (Any): Encoder hidden states.
                Defaults to None.
            encoder_attention_mask (Any): Encoder attention mask.
                Defaults to None.
            Q (Any): Q. Defaults to None.
            e1_e2_start (Any): Start and end indices for the entities in
                the input ids. Defaults to None.
            pooled_output (Any): The pooled output. Defaults to None.

        Returns:
            Optional[tuple[torch.Tensor, torch.Tensor]]:
                Logits for the relation classification task.
        """
        return None

    def output2logits(self, pooled_output: torch.Tensor,
                      sequence_output: torch.Tensor,
                      input_ids: torch.Tensor,
                      e1_e2_start: torch.Tensor) -> Optional[torch.Tensor]:
        """ Convert the output of the model to logits

        Args:
            pooled_output (torch.Tensor): output of the pooled layer.
            sequence_output (torch.Tensor): output of the sequence layer.
            input_ids (torch.Tensor): input token ids.
            e1_e2_start (torch.Tensor): start and end indices for the entities
                in the input ids.

        Returns:
            logits (torch.Tensor): logits for the relation classification task.
        """
        return None


class RelExtrBaseModel(BaseModelBluePrint):

    name = "basemodel_relcat"

    def __init__(self, relcat_config: ConfigRelCAT,
                 model_config: RelExtrBaseConfig,
                 pretrained_model_name_or_path):
        super().__init__(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            relcat_config=relcat_config,
            model_config=model_config)

        self.relcat_config: ConfigRelCAT = relcat_config
        self.model_config: RelExtrBaseConfig = model_config
        self.hf_model = PreTrainedModel(config=model_config.hf_model_config)
        self.pretrained_model_name_or_path: str = pretrained_model_name_or_path

        self._reinitialize_dense_and_frozen_layers(relcat_config=relcat_config)

        logger.info("RelCAT model config: %s",
                    str(self.model_config.hf_model_config))

    def _reinitialize_dense_and_frozen_layers(
            self, relcat_config: ConfigRelCAT) -> None:
        """ Reinitialize the dense layers of the model

        Args:
            relcat_config (ConfigRelCAT): relcat config.
        """

        self.drop_out = nn.Dropout(relcat_config.model.dropout)
        self.fc1, self.fc2, self.fc3 = create_dense_layers(relcat_config)

        for param in self.hf_model.parameters():
            if self.relcat_config.model.freeze_layers:
                param.requires_grad = False
            else:
                param.requires_grad = True

    def forward(self,
                input_ids: Optional[torch.Tensor] = None,
                attention_mask: Optional[torch.Tensor] = None,
                token_type_ids: Optional[torch.Tensor] = None,
                position_ids: Any = None,
                head_mask: Any = None,
                encoder_hidden_states: Any = None,
                encoder_attention_mask: Any = None,
                Q: Any = None,
                e1_e2_start: Any = None,
                pooled_output: Any = None
                ) -> tuple[torch.Tensor, torch.Tensor]:

        if input_ids is not None:
            input_shape = input_ids.size()
        else:
            raise ValueError("You have to specify input_ids")

        if attention_mask is None:
            attention_mask = torch.ones(
                input_shape, device=self.relcat_config.general.device)
        if encoder_attention_mask is None:
            encoder_attention_mask = torch.ones(
                input_shape, device=self.relcat_config.general.device)
        if token_type_ids is None:
            token_type_ids = torch.zeros(
                input_shape, dtype=torch.long,
                device=self.relcat_config.general.device)

        input_ids = input_ids.to(self.relcat_config.general.device)
        attention_mask = attention_mask.to(self.relcat_config.general.device)
        encoder_attention_mask = encoder_attention_mask.to(
            self.relcat_config.general.device)

        # NOTE: the wrapping of the method means that mypy can't
        #       properly understand it
        self.hf_model = self.hf_model.to(
            self.relcat_config.general.device)  # type: ignore

        model_output = self.hf_model(
            input_ids=input_ids, attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask)

        # (batch_size, sequence_length, hidden_size)
        sequence_output = model_output[0]
        pooled_output = model_output[1]

        classification_logits = self.output2logits(
            pooled_output, sequence_output, input_ids, e1_e2_start)

        return model_output, classification_logits.to(
            self.relcat_config.general.device)

    def output2logits(self, pooled_output: torch.Tensor,
                      sequence_output: torch.Tensor, input_ids: torch.Tensor,
                      e1_e2_start: torch.Tensor) -> torch.Tensor:
        """

        Args:
            pooled_output (torch.Tensor): embedding of the CLS token
            sequence_output (torch.Tensor): hidden states/embeddings for
                each token in the input text
            input_ids (torch.Tensor): input token ids.
            e1_e2_start (torch.Tensor): annotation tags token position

        Returns:
            torch.Tensor: classification probabilities for each token.
        """

        new_pooled_output = pooled_output

        if self.relcat_config.general.annotation_schema_tag_ids:
            rel_range = range(
                0,
                len(self.relcat_config.general.annotation_schema_tag_ids), 2)
            annotation_schema_tag_ids_ = [
                self.relcat_config.general.annotation_schema_tag_ids[i:i + 2]
                for i in rel_range]
            seq_tags = []

            # for each pair of tags (e1,s1) and (e2,s2)
            for each_tags in annotation_schema_tag_ids_:
                seq_tags.append(get_annotation_schema_tag(
                    sequence_output, input_ids, each_tags))

            stacked_tensor = torch.stack(seq_tags, dim=0)

            new_pooled_output = torch.cat(
                (pooled_output, *stacked_tensor), dim=1)
        else:
            e1e2_output = []
            temp_e1 = []
            temp_e2 = []

            for i, seq in enumerate(sequence_output):
                # e1e2 token sequences
                temp_e1.append(seq[e1_e2_start[i][0]])
                temp_e2.append(seq[e1_e2_start[i][1]])

            e1e2_output.append(torch.stack(temp_e1, dim=0))
            e1e2_output.append(torch.stack(temp_e2, dim=0))

            new_pooled_output = torch.cat((pooled_output, *e1e2_output), dim=1)

            del e1e2_output
            del temp_e2
            del temp_e1

        x = self.drop_out(new_pooled_output)
        x = self.fc1(x)
        x = self.drop_out(x)
        x = self.fc2(x)
        classification_logits = self.fc3(x)

        return classification_logits.to(self.relcat_config.general.device)

    @classmethod
    def load(cls, pretrained_model_name_or_path: str,
             relcat_config: ConfigRelCAT,
             model_config: RelExtrBaseConfig
             ) -> "RelExtrBaseModel":
        """ Load the model from the given path

        Args:
            pretrained_model_name_or_path (str): path to load the model from.
            relcat_config (ConfigRelCAT): relcat config.
            model_config (RelExtrBaseConfig):
                The model-specific config.

        returns:
            RelExtrBaseModel: The loaded model.
        """

        model = RelExtrBaseModel(
            relcat_config=relcat_config, model_config=model_config,
            pretrained_model_name_or_path=pretrained_model_name_or_path)

        if "modern-bert" in relcat_config.general.tokenizer_name or \
                "modern-bert" in relcat_config.general.model_name:
            from medcat.components.addons.relation_extraction.modernbert.model import RelExtrModernBertModel  # noqa
            from medcat.components.addons.relation_extraction.modernbert.config import RelExtrModernBertConfig  # noqa
            model = RelExtrModernBertModel.load_specific(
                pretrained_model_name_or_path, relcat_config=relcat_config,
                model_config=cast(RelExtrModernBertConfig, model_config))
        elif "bert" in relcat_config.general.tokenizer_name or \
             "bert" in relcat_config.general.model_name:
            from medcat.components.addons.relation_extraction.bert.model import RelExtrBertModel # noqa
            from medcat.components.addons.relation_extraction.bert.config import RelExtrBertConfig  # noqa
            model = RelExtrBertModel.load_specific(
                pretrained_model_name_or_path, relcat_config=relcat_config,
                model_config=cast(RelExtrBertConfig, model_config))
        elif "llama" in relcat_config.general.tokenizer_name or \
             "llama" in relcat_config.general.model_name:
            from medcat.components.addons.relation_extraction.llama.model import RelExtrLlamaModel # noqa
            from medcat.components.addons.relation_extraction.llama.config import RelExtrLlamaConfig # noqa
            model = RelExtrLlamaModel.load_specific(
                pretrained_model_name_or_path, relcat_config=relcat_config,
                model_config=cast(RelExtrLlamaConfig, model_config))
        else:
            if pretrained_model_name_or_path:
                model.hf_model = PreTrainedModel.from_pretrained(
                    pretrained_model_name_or_path=pretrained_model_name_or_path, config=model_config)  # noqa
            else:
                model_name = relcat_config.general.model_name
                model.hf_model = PreTrainedModel.from_pretrained(
                    pretrained_model_name_or_path=model_name,
                    config=model_config)
                logger.info("Loaded model from relcat_config: %s",
                            relcat_config.general.model_name)

        logger.info("Loaded %s from pretrained_model_name_or_path: %s",
                    str(model.__class__.__name__),
                    pretrained_model_name_or_path)

        model._reinitialize_dense_and_frozen_layers(
            relcat_config=relcat_config)

        return model



from typing import Optional, Iterable, Callable, cast, Any, Union

from abc import ABC, abstractmethod
from enum import Enum, auto
from copy import deepcopy
from pydantic import BaseModel
from itertools import islice

import numpy as np

from medcat.cat import CAT
from medcat.utils.cdb_state import captured_state_cdb

from medcat.stats.stats import get_stats
from medcat.data.mctexport import (
    MedCATTrainerExport, MedCATTrainerExportProject,
    MedCATTrainerExportDocument, MedCATTrainerExportAnnotation,
    count_all_annotations, count_all_docs, get_nr_of_annotations, iter_anns,
    iter_docs, MedCATTrainerExportProjectInfo)


class SplitType(Enum):
    """The split type."""
    DOCUMENTS = auto()
    """Split over number of documents."""
    ANNOTATIONS = auto()
    """Split over number of annotations."""
    DOCUMENTS_WEIGHTED = auto()
    """Split over number of documents based on the number of annotations.
    So essentially this ensures that the same document isn't in 2 folds
    while trying to more equally distribute documents with different number
    of annotations.
    For example:
        If we have 6 documents that we want to split into 3 folds.
        The number of annotations per document are as follows:
           [40, 40, 20, 10, 5, 5]
        If we were to split this trivially over documents, we'd end up
        with the 3 folds with number of annotations that are far from even:
           [80, 30, 10]
        However, if we use the annotations as weights, we would be able to
        create folds that have more evenly distributed annotations, e.g:
           [[D1,], [D2], [D3, D4, D5, D6]]
        where D# denotes the number of the documents, with the number of
        annotations being equal:
           [ 40, 40, 20 + 10 + 5 + 5 = 40]
    """


class FoldCreator(ABC):
    """The FoldCreator based on a MCT export.

    Args:
        mct_export (MedCATTrainerExport): The MCT export dict.
        nr_of_folds (int): Number of folds to create.
        use_annotations (bool): Whether to fold on number of annotations or
            documents.
    """

    def __init__(self, mct_export: MedCATTrainerExport, nr_of_folds: int
                 ) -> None:
        self.mct_export = mct_export
        self.nr_of_folds = nr_of_folds

    def _find_or_add_doc(self, project: MedCATTrainerExportProject,
                         orig_doc: MedCATTrainerExportDocument
                         ) -> MedCATTrainerExportDocument:
        for existing_doc in project['documents']:
            if existing_doc['name'] == orig_doc['name']:
                return existing_doc
        new_doc: MedCATTrainerExportDocument = deepcopy(orig_doc)
        new_doc['annotations'].clear()
        project['documents'].append(new_doc)
        return new_doc

    def _create_new_project(self, proj_info: MedCATTrainerExportProjectInfo
                            ) -> MedCATTrainerExportProject:
        (proj_name, proj_id, proj_cuis, proj_tuis) = proj_info
        cur_project = cast(MedCATTrainerExportProject, {
            'name': proj_name,
            'id': proj_id,
            'cuis': proj_cuis,
            'documents': [],
            })
        # NOTE: Some MCT exports don't declare TUIs
        if proj_tuis is not None:
            cur_project['tuis'] = proj_tuis
        return cur_project

    def _create_export_with_documents(self, relevant_docs: Iterable[
        tuple[MedCATTrainerExportProjectInfo, MedCATTrainerExportDocument]]
            ) -> MedCATTrainerExport:
        export: MedCATTrainerExport = {
            "projects": []
        }
        # helper for finding projects per name
        used_projects: dict[str, MedCATTrainerExportProject] = {}
        for proj_info, doc in relevant_docs:
            proj_name = proj_info[0]
            if proj_name not in used_projects:
                # TODO - make sure it's available
                cur_project = self._create_new_project(proj_info)
                export['projects'].append(cur_project)
                used_projects[proj_name] = cur_project
            else:
                cur_project = used_projects[proj_name]
            cur_project['documents'].append(doc)
        return export

    @abstractmethod
    def create_folds(self) -> list[MedCATTrainerExport]:
        """Create folds.

        Raises:
            ValueError: If something went wrong.

        Returns:
            list[MedCATTrainerExport]: The created folds.
        """


class SimpleFoldCreator(FoldCreator):

    def __init__(self, mct_export: MedCATTrainerExport, nr_of_folds: int,
                 counter: Callable[[MedCATTrainerExport], int]) -> None:
        super().__init__(mct_export, nr_of_folds)
        self._counter = counter
        self.total = self._counter(mct_export)
        self.per_fold = self._init_per_fold()

    def _init_per_fold(self) -> list[int]:
        per_fold = [self.total // self.nr_of_folds
                    for _ in range(self.nr_of_folds)]
        total = sum(per_fold)
        if total < self.total:
            per_fold[-1] += self.total - total
        if any(pf <= 0 for pf in per_fold):
            raise ValueError(
                f"Failed to calculate per-fold items. Got: {per_fold}")
        return per_fold

    @abstractmethod
    def _create_fold(self, fold_nr: int) -> MedCATTrainerExport:
        pass

    def create_folds(self) -> list[MedCATTrainerExport]:
        return [
            self._create_fold(fold_nr) for fold_nr in range(self.nr_of_folds)
        ]


class PerDocsFoldCreator(FoldCreator):

    def __init__(self, mct_export: MedCATTrainerExport, nr_of_folds: int
                 ) -> None:
        super().__init__(mct_export, nr_of_folds)
        self.nr_of_docs = count_all_docs(self.mct_export)
        self.per_doc_simple = self.nr_of_docs // self.nr_of_folds
        self._all_docs = list(iter_docs(self.mct_export))

    def _create_fold(self, fold_nr: int) -> MedCATTrainerExport:
        start_nr = self.per_doc_simple * fold_nr
        # until the end for last fold, otherwise just the next set of docs
        end_nr = (self.nr_of_docs
                  if fold_nr == self.nr_of_folds - 1
                  else start_nr + self.per_doc_simple)
        relevant_docs = self._all_docs[start_nr: end_nr]
        return self._create_export_with_documents(relevant_docs)

    def create_folds(self) -> list[MedCATTrainerExport]:
        return [
            self._create_fold(fold_nr) for fold_nr in range(self.nr_of_folds)
        ]


class PerAnnsFoldCreator(SimpleFoldCreator):

    def __init__(self, mct_export: MedCATTrainerExport, nr_of_folds: int
                 ) -> None:
        super().__init__(mct_export, nr_of_folds, count_all_annotations)

    def _add_target_ann(self, project: MedCATTrainerExportProject,
                        orig_doc: MedCATTrainerExportDocument,
                        ann: MedCATTrainerExportAnnotation) -> None:
        cur_doc: MedCATTrainerExportDocument = self._find_or_add_doc(
            project, orig_doc)
        cur_doc['annotations'].append(ann)

    def _targets(self, start_at: int
                 ) -> Iterable[tuple[MedCATTrainerExportProjectInfo,
                                     MedCATTrainerExportDocument,
                                     MedCATTrainerExportAnnotation]]:
        return islice(iter_anns(self.mct_export), start_at, None)

    def _create_fold(self, fold_nr: int) -> MedCATTrainerExport:
        per_fold = self.per_fold[fold_nr]
        already_done = sum(self.per_fold[fn] for fn in range(fold_nr))
        cur_fold: MedCATTrainerExport = {
            'projects': []
        }
        cur_project: Optional[MedCATTrainerExportProject] = None
        included = 0
        for target in self._targets(already_done):
            proj_info, cur_doc, cur_ann = target
            proj_name = proj_info[0]
            if not cur_project or cur_project['name'] != proj_name:
                # first or new project
                cur_project = self._create_new_project(proj_info)
                cur_fold['projects'].append(cur_project)
            self._add_target_ann(cur_project, cur_doc, cur_ann)
            included += 1
            if included == per_fold:
                break
            if included > per_fold:
                raise ValueError("Got a larger fold than expected. "
                                 f"Expected {per_fold}, got {included}")
        return cur_fold


class WeightedDocumentsCreator(FoldCreator):

    def __init__(self, mct_export: MedCATTrainerExport, nr_of_folds: int,
                 weight_calculator: Callable[
                     [MedCATTrainerExportDocument], int]) -> None:
        super().__init__(mct_export, nr_of_folds)
        self._weight_calculator = weight_calculator
        docs = [(doc, self._weight_calculator(doc[1]))
                for doc in iter_docs(self.mct_export)]
        # descending order in weight
        self._weighted_docs = sorted(docs, key=lambda d: d[1], reverse=True)

    def create_folds(self) -> list[MedCATTrainerExport]:
        doc_folds: list[list[tuple[MedCATTrainerExportProjectInfo,
                                   MedCATTrainerExportDocument]]]
        doc_folds = [[] for _ in range(self.nr_of_folds)]
        fold_weights = [0] * self.nr_of_folds

        for item, weight in self._weighted_docs:
            # Find the subset with the minimum total weight
            min_subset_idx = np.argmin(fold_weights)
            # add the most heavily weighted document
            doc_folds[min_subset_idx].append(item)
            fold_weights[min_subset_idx] += weight

        return [self._create_export_with_documents(docs) for docs in doc_folds]


def get_fold_creator(mct_export: MedCATTrainerExport,
                     nr_of_folds: int,
                     split_type: SplitType) -> FoldCreator:
    """Get the appropriate fold creator.

    Args:
        mct_export (MedCATTrainerExport): The MCT export.
        nr_of_folds (int): Number of folds to use.
        split_type (SplitType): The type of split to use.

    Raises:
        ValueError: In case of an unknown split type.

    Returns:
        FoldCreator: The corresponding fold creator.
    """
    if split_type is SplitType.DOCUMENTS:
        return PerDocsFoldCreator(
            mct_export=mct_export, nr_of_folds=nr_of_folds)
    elif split_type is SplitType.ANNOTATIONS:
        return PerAnnsFoldCreator(
            mct_export=mct_export, nr_of_folds=nr_of_folds)
    elif split_type is SplitType.DOCUMENTS_WEIGHTED:
        return WeightedDocumentsCreator(
            mct_export=mct_export, nr_of_folds=nr_of_folds,
            weight_calculator=get_nr_of_annotations)
    else:
        raise ValueError(f"Unknown Split Type: {split_type}")


def get_per_fold_metrics(cat: CAT, folds: list[MedCATTrainerExport],
                         use_project_filters: bool,
                         *args, **kwargs) -> list[tuple]:
    """Get per fold metrics for a given set of folds.

    This method captures the state of the before processing each fold.
    For each fold, it trains on all other folds, and runs metrics on
    the fold itself.

    Args:
        cat (CAT): The model pack.
        folds (list[MedCATTrainerExport]): The folds.
        use_project_filters (bool): Whether to use project filters.

    Returns:
        list[tuple]: The metrics for each fold.
    """
    metrics = []
    for fold_nr, cur_fold in enumerate(folds):
        others = list(folds)
        others.pop(fold_nr)
        with captured_state_cdb(cat.cdb):
            for other in others:
                cat.trainer.train_supervised_raw(
                    cast(dict[str, Any], other), *args, **kwargs)
            stats = get_stats(cat, cast(MedCATTrainerExport, cur_fold),
                              use_project_filters=use_project_filters)
            metrics.append(stats)
    return metrics


def _merge_examples(all_examples: dict, cur_examples: dict) -> None:
    for ex_type, ex_dict in cur_examples.items():
        if ex_type not in all_examples:
            all_examples[ex_type] = {}
        per_type_examples = all_examples[ex_type]
        for ex_cui, cui_examples_list in ex_dict.items():
            if ex_cui not in per_type_examples:
                per_type_examples[ex_cui] = []
            per_type_examples[ex_cui].extend(cui_examples_list)


# helper types
IntValuedMetric = Union[
    dict[str, int],
    dict[str, tuple[int, float]]
]
FloatValuedMetric = Union[
    dict[str, float],
    dict[str, tuple[float, float]]
]


class PerCUIMetrics(BaseModel):
    weights: list[Union[int, float]] = []
    vals: list[Union[int, float]] = []

    def add(self, val, weight: int = 1):
        self.weights.append(weight)
        self.vals.append(val)

    def get_mean(self):
        return sum(w * v for w, v
                   in zip(self.weights, self.vals)) / sum(self.weights)

    def get_std(self):
        mean = self.get_mean()
        return (sum(w * (v - mean)**2 for w, v
                    in zip(self.weights, self.vals)) / sum(self.weights))**.5


def _add_helper(joined: list[dict[str, PerCUIMetrics]],
                single: list[dict[str, int]]) -> None:
    if len(joined) != len(single):
        raise ValueError(
            f"Incompatible number of stuff: {len(joined)} vs {len(single)}")
    for j, s in zip(joined, single):
        for k, v in s.items():
            if k not in j:
                j[k] = PerCUIMetrics()
            j[k].add(v)


def _add_weighted_helper(joined: list[dict[str, PerCUIMetrics]],
                         single: list[dict[str, float]],
                         cui2count: dict[str, int]) -> None:
    for j, s in zip(joined, single):
        for k, v in s.items():
            if k not in j:
                j[k] = PerCUIMetrics()
            j[k].add(v, cui2count[k])


def get_metrics_mean(metrics: list[tuple[dict, dict, dict,
                                         dict, dict, dict, dict, dict]],
                     include_std: bool) -> tuple[dict, dict, dict,
                                                 dict, dict, dict, dict, dict]:
    """The the mean of the provided metrics.

    Args:
        metrics (list[tuple[dict, dict, dict, dict, dict, dict, dict, dict]):
            The metrics.
        include_std (bool): Whether to include the standard deviation.

    Returns:
        fps (dict):
            False positives for each CUI.
        fns (dict):
            False negatives for each CUI.
        tps (dict):
            True positives for each CUI.
        cui_prec (dict):
            Precision for each CUI.
        cui_rec (dict):
            Recall for each CUI.
        cui_f1 (dict):
            F1 for each CUI.
        cui_counts (dict):
            Number of occurrence for each CUI.
        examples (dict):
            Examples for each of the fp, fn, tp.
            Format will be examples['fp']['cui'][<list_of_examples>].
    """
    # additives
    all_fps: dict[str, PerCUIMetrics] = {}
    all_fns: dict[str, PerCUIMetrics] = {}
    all_tps: dict[str, PerCUIMetrics] = {}
    # weighted-averages
    all_cui_prec: dict[str, PerCUIMetrics] = {}
    all_cui_rec: dict[str, PerCUIMetrics] = {}
    all_cui_f1: dict[str, PerCUIMetrics] = {}
    # additive
    all_cui_counts: dict[str, PerCUIMetrics] = {}
    # combined
    all_additives = [
        all_fps, all_fns, all_tps, all_cui_counts
    ]
    all_weighted_averages = [
        all_cui_prec, all_cui_rec, all_cui_f1
    ]
    # examples
    all_examples: dict = {}
    for current in metrics:
        cur_wa: list = list(current[3:-2])
        cur_counts = current[-2]
        cur_adds = list(current[:3]) + [cur_counts]
        _add_helper(all_additives, cur_adds)
        _add_weighted_helper(all_weighted_averages, cur_wa, cur_counts)
        cur_examples = current[-1]
        _merge_examples(all_examples, cur_examples)
    # conversion from PerCUI metrics to int/float and (if needed) STD
    cui_fps: IntValuedMetric = {}
    cui_fns: IntValuedMetric = {}
    cui_tps: IntValuedMetric = {}
    cui_prec: FloatValuedMetric = {}
    cui_rec: FloatValuedMetric = {}
    cui_f1: FloatValuedMetric = {}
    final_counts: IntValuedMetric = {}
    to_change: list[Union[IntValuedMetric, FloatValuedMetric]] = [
        cui_fps, cui_fns, cui_tps, final_counts,
        cui_prec, cui_rec, cui_f1,
    ]
    # get the mean and/or std
    for nr, (df, d) in enumerate(zip(to_change, all_additives +
                                     all_weighted_averages)):
        for k, v in d.items():
            if nr == 3 and not include_std:
                # counts need to be added up
                # NOTE: the type:ignore comment _shouldn't_ be necessary
                #       but mypy thinks we're setting a float or integer
                #       where a tuple is expected
                df[k] = sum(v.vals)  # type: ignore
                # NOTE: The current implementation shows the sum for counts
                #       if not STD is required, but the mean along with the
                #       standard deviation if the latter is required.
            elif not include_std:
                df[k] = v.get_mean()
            else:
                # NOTE: the type:ignore comment _shouldn't_ be necessary
                #       but mypy thinks we're setting a tuple
                #       where a float or integer is expected
                df[k] = (v.get_mean(), v.get_std())  # type: ignore
    return (cui_fps, cui_fns, cui_tps, cui_prec, cui_rec, cui_f1,
            final_counts, all_examples)


def get_k_fold_stats(cat: CAT, mct_export_data: MedCATTrainerExport,
                     k: int = 3, use_project_filters: bool = False,
                     split_type: SplitType = SplitType.DOCUMENTS_WEIGHTED,
                     include_std: bool = False, *args, **kwargs) -> tuple:
    """Get the k-fold stats for the model with the specified data.

    First this will split the MCT export into `k` folds. You can do
    this either per document or per-annotation.

    For each of the `k` folds, it will start from the base model,
    train it with with the other `k-1` folds and record the metrics.
    After that the base model state is restored before doing the next fold.
    After all the folds have been done, the metrics are averaged.

    Args:
        cat (CAT): The model pack.
        mct_export_data (MedCATTrainerExport): The MCT export.
        k (int): The number of folds. Defaults to 3.
        use_project_filters (bool): Whether to use per project filters.
            Defaults to `False`.
        split_type (SplitType): Whether to use annodations or docs.
            Defaults to DOCUMENTS_WEIGHTED.
        include_std (bool): Whether to include stanrdard deviation.
            Defaults to False.
        *args: Arguments passed to the `CAT.train_supervised_raw` method.
        **kwargs: Keyword arguments passed to the `CAT.train_supervised_raw`
            method.

    Returns:
        tuple: The averaged metrics. Potentially with their corresponding
            standard deviations.
    """
    creator = get_fold_creator(mct_export_data, k, split_type=split_type)
    folds = creator.create_folds()
    per_fold_metrics = get_per_fold_metrics(
        cat, folds, use_project_filters, *args, **kwargs)
    means = get_metrics_mean(per_fold_metrics, include_std)
    return means



from typing import Optional

from medcat.components.types import CoreComponentType, AbstractCoreComponent
from medcat.tokenizing.tokens import MutableDocument
from medcat.tokenizing.tokenizers import BaseTokenizer
from medcat.cdb.cdb import CDB
from medcat.vocab import Vocab
from medcat.config.config import ComponentConfig


class NoActionLinker(AbstractCoreComponent):
    name = 'no_action'

    def get_type(self):
        return CoreComponentType.linking

    def __call__(self, doc: MutableDocument) -> MutableDocument:
        return doc

    @classmethod
    def create_new_component(
            cls, cnf: ComponentConfig, tokenizer: BaseTokenizer,
            cdb: CDB, vocab: Vocab, model_load_path: Optional[str]
            ) -> 'NoActionLinker':
        return cls()



from typing import Optional, Callable, cast, Type
import re
import os
import shutil
import logging

import spacy
from spacy.tokens import Span
from spacy.tokenizer import Tokenizer  # type: ignore
from spacy.language import Language

from medcat.tokenizing.tokens import (MutableDocument, MutableEntity,
                                      MutableToken)
from medcat.tokenizing.tokenizers import BaseTokenizer, TOKENIZER_PREFIX
from medcat.tokenizing.spacy_impl.tokens import Document, Entity, Token
from medcat.tokenizing.spacy_impl.utils import ensure_spacy_model
from medcat.config import Config


logger = logging.getLogger(__name__)


def spacy_split_all(nlp: Language, use_diacritics: bool) -> Tokenizer:

    token_characters = r'[^A-Za-z0-9\@]'

    if use_diacritics:
        token_characters = r'[^A-Za-zÀ-ÖØ-öø-ÿ0-9\@]'

    infix_re = re.compile(token_characters)
    suffix_re = re.compile(token_characters + r'$')
    prefix_re = re.compile(r'^' + token_characters)
    return Tokenizer(nlp.vocab,
                     rules={},
                     token_match=None,
                     prefix_search=prefix_re.search,
                     suffix_search=suffix_re.search,
                     infix_finditer=infix_re.finditer
                     )


class SpacyTokenizer(BaseTokenizer):

    def __init__(self, spacy_model_name: str,
                 spacy_disabled_components: list[str],
                 use_diacritics: bool,
                 max_document_length: int,
                 tokenizer_getter: Callable[[Language, bool], Tokenizer
                                            ] = spacy_split_all,
                 stopwords: Optional[set[str]] = None,):
        self._spacy_model_name = os.path.basename(
            spacy_model_name).removeprefix(TOKENIZER_PREFIX)
        if self.load_internals_from(spacy_model_name):
            # i.e has something to load from path
            pass
        else:
            # no file provided, ensure the model is available
            ensure_spacy_model(self._spacy_model_name)
            spacy_model_name = self._spacy_model_name
        if stopwords is not None:
            lang_str = os.path.basename(spacy_model_name).split('_', 1)[0]
            cls = spacy.util.get_lang_class(lang_str)
            cls.Defaults.stop_words = set(stopwords)
        self._nlp = spacy.load(spacy_model_name,
                               disable=spacy_disabled_components)
        self._nlp.tokenizer = tokenizer_getter(self._nlp, use_diacritics)
        self._nlp.max_length = max_document_length

    def create_entity(self, doc: MutableDocument,
                      token_start_index: int, token_end_index: int,
                      label: str) -> MutableEntity:
        spacy_doc = cast(Document, doc)._delegate
        span = Span(spacy_doc, token_start_index, token_end_index, label)
        return Entity(span)

    def entity_from_tokens(self, tokens: list[MutableToken]) -> MutableEntity:
        if not tokens:
            raise ValueError("Need at least one token for an entity")
        spacy_tokens = cast(list[Token], tokens)
        span = Span(spacy_tokens[0]._delegate.doc, spacy_tokens[0].index,
                    spacy_tokens[-1].index + 1)
        return Entity(span)

    def __call__(self, text: str) -> MutableDocument:
        return Document(self._nlp(text))

    @classmethod
    def create_new_tokenizer(cls, config: Config) -> 'SpacyTokenizer':
        nlp_cnf = config.general.nlp
        return cls(nlp_cnf.modelname,
            nlp_cnf.disabled_components,
            config.general.diacritics,
            config.preprocessing.max_document_length,
            stopwords=config.preprocessing.stopwords)

    def get_doc_class(self) -> Type[MutableDocument]:
        return Document

    def get_entity_class(self) -> Type[MutableEntity]:
        return Entity

    # saveable tokenizer

    def save_internals_to(self, folder_path: str) -> str:
        subfolder = os.path.join(
            folder_path, f"{TOKENIZER_PREFIX}{self._spacy_model_name}")
        if os.path.exists(subfolder):
            # NOTE: always overwrite
            shutil.rmtree(folder_path)
        logger.debug("Saving spacy model to '%s'", subfolder)
        cur_path = self._nlp._path
        if cur_path is None:
            raise ValueError(f"Unable to save spacy: {self._nlp}")
        shutil.copytree(cur_path, subfolder)
        return subfolder

    def load_internals_from(self, folder_path: str) -> bool:
        return os.path.exists(folder_path)



import os
import json
import logging
import numpy
from multiprocessing import Lock
from datetime import datetime
from typing import Iterable, Optional, cast, Union, Any, TypedDict, Callable

from medcat.utils.hasher import Hasher

import torch
from torch import nn, Tensor
from medcat.tokenizing.tokenizers import BaseTokenizer
from medcat.config.config import ComponentConfig
from medcat.config.config_meta_cat import ConfigMetaCAT
from medcat.components.addons.meta_cat.ml_utils import (
    predict, train_model, set_all_seeds, eval_model)
from medcat.components.addons.meta_cat.data_utils import (
    prepare_from_json, encode_category_values, prepare_for_oversampled_data)
from medcat.components.addons.addons import AddonComponent
from medcat.components.addons.meta_cat.mctokenizers.tokenizers import (
    TokenizerWrapperBase, init_tokenizer, load_tokenizer)
from medcat.storage.serialisers import serialise, deserialise
from medcat.storage.serialisables import (
    AbstractSerialisable, SerialisingStrategy)
from medcat.tokenizing.tokens import MutableDocument, MutableEntity
from medcat.cdb import CDB
from medcat.vocab import Vocab
from medcat.utils.defaults import COMPONENTS_FOLDER
from medcat.utils.defaults import (
    avoid_legacy_conversion, doing_legacy_conversion_message,
    LegacyConversionDisabledError)
from peft import get_peft_model, LoraConfig, TaskType

# It should be safe to do this always, as all other multiprocessing
# will be finished before data comes to meta_cat
os.environ["TOKENIZERS_PARALLELISM"] = "true"

logger = logging.getLogger(__name__)


_META_ANNS_PATH = 'meta_cat_meta_anns'
_SHARE_TOKENS_PATH = 'meta_cat_share_tokens'


class MedCATTrainerExportDocument(TypedDict):
    name: str
    confidence: float
    value: str


class MetaAnnotationValue(TypedDict):
    name: str
    value: str
    confidence: float


TokenizerPreprocessor = Optional[
    Callable[[Optional[TokenizerWrapperBase]], None]]


class MetaCATAddon(AddonComponent):
    DEFAULT_TOKENIZER = 'spacy'
    addon_type = 'meta_cat'
    output_key = 'meta_anns'
    config: ConfigMetaCAT

    def __init__(self, config: ConfigMetaCAT, base_tokenizer: BaseTokenizer,
                 meta_cat: Optional['MetaCAT']) -> None:
        self.config = config
        self._mc = meta_cat
        self._name = config.general.category_name
        self._init_data_paths(base_tokenizer)

    @property
    def mc(self) -> 'MetaCAT':
        if self._mc is None:
            raise ValueError("Need to have specified MetaCAT")
        return self._mc

    @classmethod
    def create_new(cls, config: ConfigMetaCAT, base_tokenizer: BaseTokenizer,
                   tknzer_preprocessor: TokenizerPreprocessor = None
                   ) -> 'MetaCATAddon':
        """Factory method to create a new MetaCATAddon instance."""
        tokenizer = init_tokenizer(config)
        if tknzer_preprocessor is not None:
            tknzer_preprocessor(tokenizer)
        meta_cat = MetaCAT(tokenizer, embeddings=None, config=config)
        return cls(config, base_tokenizer, meta_cat)

    @classmethod
    def create_new_component(
            cls, cnf: ComponentConfig, tokenizer: BaseTokenizer,
            cdb: CDB, vocab: Vocab, model_load_path: Optional[str]
            ) -> 'MetaCATAddon':
        if not isinstance(cnf, ConfigMetaCAT):
            raise ValueError(f"Incompatible config: {cnf}")
        if model_load_path is not None:
            components_folder = os.path.join(
                model_load_path, COMPONENTS_FOLDER)
            folder_name = cls.get_folder_name_for_addon_and_name(
                cls.addon_type, str(cnf.general.category_name))
            load_path = os.path.join(components_folder, folder_name)
            return cls.load_existing(cnf, tokenizer, load_path)
        # TODO: tokenizer preprocessing for (e.g) BPE tokenizer (see PR #67)
        return cls.create_new(cnf, tokenizer, None)

    @classmethod
    def load_existing(cls, cnf: ConfigMetaCAT,
                      base_tokenizer: BaseTokenizer,
                      load_path: str) -> 'MetaCATAddon':
        """Factory method to load an existing MetaCATAddon from disk."""
        meta_cat = cls(cnf, base_tokenizer, None)  # Temporary instance
        meta_cat._mc = meta_cat.load(load_path)
        return meta_cat

    @property
    def name(self) -> str:
        return str(self._name)

    def __call__(self, doc: MutableDocument) -> MutableDocument:
        return self.mc(doc)

    def load(self, folder_path: str) -> 'MetaCAT':
        mc_path, tokenizer_folder = self._get_meta_cat_and_tokenizer_paths(
            folder_path)
        mc = cast(MetaCAT, deserialise(mc_path))
        mc.tokenizer = self._load_tokenizer(self.config, tokenizer_folder)
        return mc

    @classmethod
    def _load_tokenizer(cls, config: ConfigMetaCAT, tokenizer_folder: str
                        ) -> Optional[TokenizerWrapperBase]:
        return load_tokenizer(config, tokenizer_folder)

    @classmethod
    def _get_meta_cat_and_tokenizer_paths(cls, folder_path: str
                                          ) -> tuple[str, str]:
        return (os.path.join(folder_path, 'meta_cat'),
                os.path.join(folder_path, "tokenizer"))

    def save(self, folder_path: str) -> None:
        mc_path, tokenizer_folder = self._get_meta_cat_and_tokenizer_paths(
            folder_path)
        os.mkdir(mc_path)
        os.mkdir(tokenizer_folder)
        serialise(self.config.general.serialiser, self.mc, mc_path)
        if self.mc.tokenizer is None:
            raise MisconfiguredMetaCATException(
                "Unable to save MetaCAT without a tokenizer")
        self.mc.tokenizer.save(tokenizer_folder)

    def _init_data_paths(self, base_tokenizer: BaseTokenizer):
        # a dictionary like {category_name: value, ...}
        base_tokenizer.get_entity_class().register_addon_path(
            _META_ANNS_PATH, def_val=None, force=True)
        # Used for sharing pre-processed data/tokens
        base_tokenizer.get_doc_class().register_addon_path(
            _SHARE_TOKENS_PATH, def_val=None, force=True)

    @property
    def include_in_output(self) -> bool:
        return True

    def get_output_key_val(self, ent: MutableEntity
                           ) -> tuple[str, dict[str, MetaAnnotationValue]]:
        # NOTE: In case of multiple MetaCATs, this will be called
        #       once for each MetaCAT and will get the same value.
        #       But it shouldn't be too much of an issue.
        return self.output_key, ent.get_addon_data(_META_ANNS_PATH)

    # for ManualSerialisable:

    def serialise_to(self, folder_path: str) -> None:
        os.mkdir(folder_path)
        self.save(folder_path)

    @classmethod
    def _create_throwaway_tokenizer(cls) -> BaseTokenizer:
        from medcat.tokenizing.tokenizers import create_tokenizer
        from medcat.config import Config
        logger.warning(
            "A base tokenizer was not provided during the loading of a "
            "MetaCAT. The tokenizer is used to register the required data "
            "paths for MetaCAT to function. Using the default of '%s'. If "
            "this it not the tokenizer you will end up using, MetaCAT may "
            "be unable to recover unless a) the paths are registered "
            "explicitly, or b) there are other MetaCATs created with the "
            "correct tokenizer. Do note that this will also create "
            "another instance of the tokenizer, though it should be "
            "garbage collected soon.", cls.DEFAULT_TOKENIZER
        )
        # NOTE: the use of a (mostly) default config here probably won't
        #       affect anything since the tokenizer itself won't be used
        gcnf = Config()
        gcnf.general.nlp.provider = 'spacy'
        return create_tokenizer(cls.DEFAULT_TOKENIZER, gcnf)

    @classmethod
    def deserialise_from(cls, folder_path: str, **init_kwargs
                         ) -> 'MetaCATAddon':
        if "model.dat" in os.listdir(folder_path):
            if not avoid_legacy_conversion():
                doing_legacy_conversion_message(
                    logger, cls.__name__, folder_path)
                from medcat.utils.legacy.convert_meta_cat import (
                    get_meta_cat_from_old)
                return get_meta_cat_from_old(
                    folder_path, cls._create_throwaway_tokenizer())
            raise LegacyConversionDisabledError(cls.__name__,)
        if 'cnf' in init_kwargs:
            cnf = init_kwargs['cnf']
        else:
            config_path = os.path.join(folder_path, "meta_cat", "config")
            if not os.path.exists(config_path):
                # load legacy config (assuming it exists)
                config_path += ".dat"
            logger.info(
                "Was not provide a config when loading a meta cat from '%s'. "
                "Inferring config from file at '%s'", folder_path,
                config_path)
            cnf = ConfigMetaCAT.load(config_path)
        if 'tokenizer' in init_kwargs:
            tokenizer = init_kwargs['tokenizer']
        else:
            tokenizer = cls._create_throwaway_tokenizer()
        return cls.load_existing(
            load_path=folder_path,
            cnf=cnf,
            base_tokenizer=tokenizer)

    def get_strategy(self) -> SerialisingStrategy:
        return SerialisingStrategy.MANUAL

    @classmethod
    def get_init_attrs(cls) -> list[str]:
        return []

    @classmethod
    def ignore_attrs(cls) -> list[str]:
        return []

    @classmethod
    def include_properties(cls) -> list[str]:
        return []

    def get_hash(self) -> str:
        if self._mc:
            return self._mc.get_hash()
        else:
            return 'No-model'


def get_meta_annotations(entity: MutableEntity
                         ) -> dict[str, MetaAnnotationValue]:
    return entity.get_addon_data(_META_ANNS_PATH)


class MetaCAT(AbstractSerialisable):
    """The MetaCAT class used for training 'Meta-Annotation' models,
    i.e. annotations of clinical concept annotations. These are also
    known as properties or attributes of recognise entities sin similar
    tools such as MetaMap and cTakes.

    This is a flexible model agnostic class that can learns any
    meta-annotation task, i.e. any multi-class classification task
    for recognised terms.

    Args:
        tokenizer (TokenizerWrapperBase):
            The Huggingface tokenizer instance. This can be a pre-trained
            tokenzier instance from a BERT-style model, or trained from
            scratch for the Bi-LSTM (w. attention) model that is currentl
              used in most deployments.
        embeddings (Tensor, numpy.ndarray):
            embedding mapping (sub)word input id n-dim (sub)word embedding.
        config (ConfigMetaCAT):
            the configuration for MetaCAT. Param descriptions available in
            ConfigMetaCAT docs.
    """

    # Custom pipeline component name
    name = 'meta_cat'
    _component_lock = Lock()

    @classmethod
    def get_init_attrs(cls) -> list[str]:
        return ['tokenizer', 'embeddings', 'config',
                '_model_state_dict']

    @classmethod
    def ignore_attrs(cls) -> list[str]:
        return ['model']

    @classmethod
    def include_properties(cls) -> list[str]:
        return ['_model_state_dict']

    @property
    def _model_state_dict(self):
        return self.model.state_dict()

    # Override
    def __init__(self,
                 tokenizer: Optional[TokenizerWrapperBase] = None,
                 embeddings: Optional[Union[Tensor, numpy.ndarray]] = None,
                 config: Optional[ConfigMetaCAT] = None,
                 _model_state_dict: Optional[dict[str, Any]] = None) -> None:
        if config is None:
            config = ConfigMetaCAT()
        self.config = config
        set_all_seeds(config.general.seed)

        self.tokenizer = tokenizer
        if tokenizer is not None:
            self._reset_tokenizer_info()

        self.embeddings = (torch.tensor(
            embeddings, dtype=torch.float32) if embeddings is not None
            else None)
        self.model = self.get_model(embeddings=self.embeddings)
        if _model_state_dict:
            self.model.load_state_dict(_model_state_dict)

    def _reset_tokenizer_info(self):
        # Set it in the config
        self.config.general.tokenizer_name = self.tokenizer.name
        self.config.general.vocab_size = self.tokenizer.get_size()
        # We will also set the padding
        self.config.model.padding_idx = cast(int, self.tokenizer.get_pad_id())

    def get_model(self, embeddings: Optional[Tensor]) -> nn.Module:
        """Get the model

        Args:
            embeddings (Optional[Tensor]):
                The embedding densor

        Raises:
            ValueError: If the meta model is not LSTM or BERT

        Returns:
            nn.Module:
                The module
        """
        config = self.config
        if config.model.model_name == 'lstm':
            from medcat.components.addons.meta_cat.models import LSTM
            model: nn.Module = LSTM(embeddings, config)
            logger.info("LSTM model used for classification")

        elif config.model.model_name == 'bert':
            from medcat.components.addons.meta_cat.models import (
                BertForMetaAnnotation)
            model = BertForMetaAnnotation(config)

            if not config.model.model_freeze_layers:
                peft_config = LoraConfig(
                    task_type=TaskType.SEQ_CLS, inference_mode=False, r=8,
                    lora_alpha=16, target_modules=["query", "value"],
                    lora_dropout=0.2)

                # NOTE: Not sure what changed between transformers 4.50.3 and
                # 4.50.1 that made this fail for mypy. But as best as I can
                # tell, it still works just the same
                model = get_peft_model(model, peft_config)  # type: ignore
                # model.print_trainable_parameters()

            logger.info("BERT model used for classification")

        else:
            raise ValueError("Unknown model name %s" % config.model.model_name)

        return model

    def get_hash(self) -> str:
        """A partial hash trying to catch differences between models.

        Returns:
            str: The hex hash.
        """
        hasher = Hasher()
        # Set last_train_on if None
        if self.config.train.last_train_on is None:
            self.config.train.last_train_on = datetime.now().timestamp()

        hasher.update(self.config.model_dump())
        return hasher.hexdigest()

    def train_from_json(self, json_path: Union[str, list],
                        save_dir_path: Optional[str] = None,
                        data_oversampled: Optional[list] = None,
                        overwrite: bool = False) -> dict:
        """Train or continue training a model give a json_path containing
        a MedCATtrainer export. It will continue training if an existing
        model is loaded or start new training if the model is blank/new.

        Args:
            json_path (Union[str, list]):
                Path/Paths to a MedCATtrainer export containing the
                meta_annotations we want to train for.
            save_dir_path (Optional[str]):
                In case we have aut_save_model (meaning during the
                training the best model will be saved) we need to
                set a save path. Defaults to `None`.
            data_oversampled (Optional[list]):
                In case of oversampling being performed, the data
                will be passed in the parameter allowing the
                model to be trained on original + synthetic data.
            overwrite (bool):
                Whether to allow overwriting the file if/when appropriate.

        Returns:
            dict: The resulting report.
        """

        # Load the medcattrainer export
        if isinstance(json_path, str):
            json_path = [json_path]

        def merge_data_loaded(base, other):
            if not base:
                return other
            elif other is None:
                return base
            else:
                for p in other['projects']:
                    base['projects'].append(p)
            return base

        # Merge data from all different data paths
        data_loaded: dict = {}
        for path in json_path:
            with open(path, 'r') as f:
                data_loaded = merge_data_loaded(data_loaded, json.load(f))
        return self.train_raw(data_loaded, save_dir_path,
                              data_oversampled=data_oversampled,
                              overwrite=overwrite)

    def train_raw(self, data_loaded: dict, save_dir_path: Optional[str] = None,
                  data_oversampled: Optional[list] = None,
                  overwrite: bool = False) -> dict:
        """
        Train or continue training a model given raw data. It will continue
        training if an existing model is loaded or start new training if
        the model is blank/new.

        The raw data is expected in the following format:
        {
            'projects': [  # list of projects
                {
                    'name': '<project_name>',
                    'documents': [  # list of documents
                        {
                            'name': '<document_name>',
                            'text': '<text_of_document>',
                            'annotations': [  # list of annotations
                                {
                                    # start index of the annotation
                                    'start': -1,
                                    'end': 1,    # end index of the annotation
                                    'cui': 'cui',
                                    'value': '<annotation_value>'
                                },
                                ...
                            ],
                        },
                        ...
                    ]
                },
                ...
            ]
        }

        Args:
            data_loaded (dict):
                The raw data we want to train for.
            save_dir_path (Optional[str]):
                In case we have aut_save_model (meaning during the training
                the best model will be saved) we need to set a save path.
                Defaults to `None`.
            data_oversampled (Optional[list]):
                In case of oversampling being performed, the data will be
                passed in the parameter allowing the model to be trained on
                original + synthetic data. The format of which is expected:
                [[['text','of','the','document'], [index of medical entity],
                    "label" ],
                ['text','of','the','document'], [index of medical entity],
                    "label" ]]
            overwrite (bool):
                Whether to allow overwriting the file if/when appropriate.

        Returns:
            dict: The resulting report.

        Raises:
            Exception: If no save path is specified, or category name
                not in data.
            AssertionError: If no tokeniser is set
            FileNotFoundError: If phase_number is set to 2 and model.dat
                file is not found
            KeyError: If phase_number is set to 2 and model.dat file
                contains mismatched architecture
        """
        g_config = self.config.general
        t_config = self.config.train

        # Create directories if they don't exist
        if t_config.auto_save_model:
            if save_dir_path is None:
                raise Exception("The `save_dir_path` argument is required if "
                                "`aut_save_model` is `True` in the config")
            else:
                os.makedirs(save_dir_path, exist_ok=True)

        # Prepare the data
        assert self.tokenizer is not None
        data_in = prepare_from_json(
            data_loaded, g_config.cntx_left, g_config.cntx_right,
            self.tokenizer, cui_filter=t_config.cui_filter,
            replace_center=g_config.replace_center,
            prerequisites=t_config.prerequisites, lowercase=g_config.lowercase)

        # Check is the name present
        category_name = category_name = g_config.get_applicable_category_name(
            data_in)
        if category_name is None:
            raise Exception(
                "The category name does not exist in this json file. "
                f"You've provided '{category_name}', while the possible "
                f"options are: {' | '.join(list(data_in.keys()))}. "
                "Additionally, ensure the populate the "
                "'alternative_category_names' attribute to accommodate "
                "for variations.")

        data = data_in[category_name]
        if data_oversampled:
            data_sampled = prepare_for_oversampled_data(
                data_oversampled, self.tokenizer)
            data = data + data_sampled

        category_value2id = g_config.category_value2id
        if not category_value2id:
            # Encode the category values
            (full_data, data_undersampled,
             category_value2id) = encode_category_values(
                 data,
                 category_undersample=self.config.model.category_undersample,
                 alternative_class_names=g_config.alternative_class_names)
        else:
            # We already have everything, just get the data
            (full_data, data_undersampled,
             category_value2id) = encode_category_values(
                 data, existing_category_value2id=category_value2id,
                 category_undersample=self.config.model.category_undersample,
                 alternative_class_names=g_config.alternative_class_names)
            g_config.category_value2id = category_value2id
            self.config.model.nclasses = len(category_value2id)
        # Make sure the config number of classes is the same
        # as the one found in the data
        if len(category_value2id) != self.config.model.nclasses:
            logger.warning(
                "The number of classes set in the config is not the same as "
                f"the one found in the data: {self.config.model.nclasses} vs "
                f"{len(category_value2id)}")
            logger.warning("Auto-setting the nclasses value in config and "
                           "rebuilding the model.")
            self.config.model.nclasses = len(category_value2id)

        if self.config.model.phase_number == 2 and save_dir_path is not None:
            model_save_path = os.path.join(save_dir_path, 'model.dat')
            device = torch.device(g_config.device)
            try:
                self.model.load_state_dict(torch.load(
                    model_save_path, map_location=device))
                logger.info("Training model for Phase 2, with model dict "
                            "loaded from disk")
            except FileNotFoundError:
                raise FileNotFoundError(
                    f"\nError: Model file not found at path: {model_save_path}"
                    "\nPlease run phase 1 training and then run phase 2.")

            except KeyError:
                raise KeyError(
                    "\nError: Missing key in loaded state dictionary. "
                    "\nThis might be due to a mismatch between the model "
                    "architecture and the saved state.")

            except Exception as e:
                raise Exception(
                    f"\nError: Model state cannot be loaded from dict. {e}")

        data = full_data
        if self.config.model.phase_number == 1:
            data = data_undersampled
            if not t_config.auto_save_model:
                logger.info("For phase 1, model state has to be saved. "
                            "Saving model...")
                t_config.auto_save_model = True
            logger.info("Training model for Phase 1 now...")

        report = train_model(self.model, data=data, config=self.config,
                             save_dir_path=save_dir_path)

        # If autosave, then load the best model here
        if t_config.auto_save_model:
            if save_dir_path is None:
                raise Exception("The `save_dir_path` argument is required if "
                                "`aut_save_model` is `True` in the config")
            else:
                path = os.path.join(save_dir_path, 'model.dat')
                device = torch.device(g_config.device)
                self.model.load_state_dict(torch.load(
                    path, map_location=device))

                # Save everything now
                serialise(self.config.general.serialiser, self, save_dir_path,
                          overwrite=overwrite)

        self.config.train.last_train_on = datetime.now().timestamp()
        return report

    def eval(self, json_path: str) -> dict:
        """Evaluate from json.

        Args:
            json_path (str):
                The json file ath

        Returns:
            dict:
                The resulting model dict

        Raises:
            AssertionError: If self.tokenizer
            Exception: If the category name does not exist
        """
        g_config = self.config.general
        t_config = self.config.train

        with open(json_path, 'r') as f:
            data_loaded: dict = json.load(f)

        # Prepare the data
        assert self.tokenizer is not None
        data_in = prepare_from_json(
            data_loaded, g_config.cntx_left, g_config.cntx_right,
            self.tokenizer, cui_filter=t_config.cui_filter,
            replace_center=g_config.replace_center,
            prerequisites=t_config.prerequisites, lowercase=g_config.lowercase)

        # Check is the name there
        category_name = g_config.get_applicable_category_name(data_in)
        if category_name is None:
            raise Exception(
                "The category name does not exist in this json file.")

        data = data_in[category_name]

        # We already have everything, just get the data
        category_value2id = g_config.category_value2id
        data, _, _ = encode_category_values(
            data, existing_category_value2id=category_value2id)

        # Run evaluation
        assert self.tokenizer is not None
        result = eval_model(self.model, data, config=self.config,
                            tokenizer=self.tokenizer)

        return result

    def get_ents(self, doc: MutableDocument) -> Iterable[MutableEntity]:
        # TODO - use span groups?
        return doc.ner_ents  # TODO: is this correct?

    def prepare_document(self, doc: MutableDocument, input_ids: list,
                         offset_mapping: list, lowercase: bool
                         ) -> tuple[dict, list]:
        """Prepares document.

        Args:
            doc (Doc):
                The document
            input_ids (list):
                Input ids
            offset_mapping (list):
                Offset mappings
            lowercase (bool):
                Whether to use lower case replace center

        Returns:
            tuple[dict, list]:
                Entity id to index mapping
                and
                Samples
        """
        config = self.config
        cntx_left = config.general.cntx_left
        cntx_right = config.general.cntx_right
        replace_center = config.general.replace_center

        ents = self.get_ents(doc)

        samples = []
        last_ind = 0
        # Map form entity ID to where is it in the samples array
        ent_id2ind = {}
        for ent in sorted(ents, key=lambda ent: ent.base.start_char_index):
            start = ent.base.start_char_index
            end = ent.base.end_char_index

            # Updated implementation to extract all the tokens for
            # the medical entity (rather than the one)
            ctoken_idx = []
            for ind, pair in enumerate(offset_mapping[last_ind:]):
                # Checking if we've reached at the start of the entity
                if start <= pair[0] or start <= pair[1]:
                    if end <= pair[1]:
                        ctoken_idx.append(ind)  # End reached
                        break
                    else:
                        ctoken_idx.append(ind)  # Keep going

            # Start where the last ent was found, cannot be before it as we've
            # sorted
            last_ind += ind  # If we did not start from 0 in the for loop

            _start = max(0, ctoken_idx[0] - cntx_left)
            _end = min(len(input_ids), ctoken_idx[-1] + 1 + cntx_right)

            tkns = input_ids[_start:_end]
            cpos = cntx_left + min(0, ind - cntx_left)
            cpos_new = [x - _start for x in ctoken_idx]

            if replace_center is not None:
                if lowercase:
                    replace_center = replace_center.lower()
                # We start from ind
                s_ind = ind
                e_ind = ind
                for _ind, pair in enumerate(offset_mapping[ind:]):
                    if end > pair[0] and end <= pair[1]:
                        e_ind = _ind + ind
                        break
                ln = e_ind - s_ind  # Length of the concept in tokens
                assert self.tokenizer is not None
                tkns = tkns[:cpos] + self.tokenizer(
                    replace_center)['input_ids'] + tkns[cpos + ln + 1:]
            samples.append([tkns, cpos_new])
            ent_id2ind[ent.id] = len(samples) - 1

        return ent_id2ind, samples

    @staticmethod
    def batch_generator(stream: Iterable[MutableDocument],
                        batch_size_chars: int
                        ) -> Iterable[list[MutableDocument]]:
        """Generator for batch of documents.

        Args:
            stream (Iterable[MutableDocument]):
                The document stream
            batch_size_chars (int):
                Number of characters per batch

        Yields:
            list[MutableDocument]: The batch of documents.
        """
        docs = []
        char_count = 0
        for doc in stream:
            char_count += len(doc.base.text)
            docs.append(doc)
            if char_count < batch_size_chars:
                continue
            yield docs
            docs = []
            char_count = 0

        # If there is anything left return that also
        if len(docs) > 0:
            yield docs

    def _set_meta_anns(self,
                       doc: MutableDocument,
                       id2category_value: dict
                       ) -> MutableDocument:
        config = self.config
        data: list
        if (not config.general.save_and_reuse_tokens or
                doc.get_addon_data(_SHARE_TOKENS_PATH) is None):
            if config.general.lowercase:
                all_text = doc.base.text.lower()
            else:
                all_text = doc.base.text
            assert self.tokenizer is not None
            all_text_processed = self.tokenizer(all_text)
            ent_id2ind, data = self.prepare_document(
                doc, input_ids=all_text_processed['input_ids'],
                offset_mapping=all_text_processed['offset_mapping'],
                lowercase=config.general.lowercase)
        else:
            # This means another model has already processed the data
            # and we can just use it. This is a
            # dangerous option - as it assumes the other model has the
            # same tokenizer and context size.
            data = []
            data.extend(doc.get_addon_data(_SHARE_TOKENS_PATH)[0])
        predictions, confidences = predict(
            self.model, data, config)

        ents = self.get_ents(doc)

        for ent in ents:
            ent_ind = ent_id2ind[ent.id]
            value = id2category_value[predictions[ent_ind]]
            confidence = confidences[ent_ind]
            if ent.get_addon_data(_META_ANNS_PATH) is None:
                ent.set_addon_data(_META_ANNS_PATH, {
                    config.general.category_name: {
                        'value': value,
                        'confidence': float(confidence),
                        'name': config.general.category_name
                    }
                }
                )
            else:
                ent.get_addon_data(_META_ANNS_PATH)[
                        config.general.category_name] = {
                    'value': value,
                    'confidence': float(confidence),
                    'name': config.general.category_name
                }
        return doc

    # Override
    def __call__(self, doc: MutableDocument) -> MutableDocument:
        """Process one document, used in the spacy pipeline for sequential
        document processing.

        Args:
            doc (Doc):
                A spacy document

        Returns:
            Doc: The same spacy document.
        """
        id2category_value = {
            v: k for k, v in self.config.general.category_value2id.items()}
        self._set_meta_anns(doc, id2category_value)
        return doc

    def get_model_card(self, as_dict: bool = False) -> Union[str, dict]:
        """A minimal model card.

        Args:
            as_dict (bool):
                Return the model card as a dictionary instead of a str.
                Defaults to `False`.

        Returns:
            Union[str, dict]:
                An indented JSON object.
                OR A JSON object in dict form.
        """
        card = {
            'Category Name': self.config.general.category_name,
            'Description': self.config.general.description,
            'Classes': self.config.general.category_value2id,
            'Model': self.config.model.model_name
        }
        if as_dict:
            return card
        else:
            return json.dumps(card, indent=2, sort_keys=False)

    def __repr__(self):
        """Prints the model_card for this MetaCAT instance.

        Returns:
            the 'Model Card' for this MetaCAT instance. This includes NER+L
            config and any MetaCATs
        """
        return self.get_model_card(as_dict=False)


class MisconfiguredMetaCATException(ValueError):

    def ____(self, *args):
        super().__init__(*args)



import os
import json
import logging

import torch

from medcat.cdb import CDB
from medcat.components.addons.relation_extraction.rel_cat import (
    RelCAT, RelCATAddon)
from medcat.components.addons.relation_extraction.base_component import (
    RelExtrBaseComponent)
from medcat.config.config_rel_cat import ConfigRelCAT
from medcat.tokenizing.tokenizers import BaseTokenizer, create_tokenizer
from medcat.utils.legacy.helpers import fix_old_style_cnf


logger = logging.getLogger(__name__)


def _load_legacy(cdb: CDB, base_tokenizer: BaseTokenizer,
                 config: ConfigRelCAT, save_dir_path: str) -> RelCAT:
    # tokenizer: Optional[TokenizerWrapperBase] = None
    # Load tokenizer
    # tokenizer = load_tokenizer(config, save_dir_path)

    # Create rel_cat
    # rel_cat = RelCAT(tokenizer=tokenizer, embeddings=None, config=config)
    component = RelExtrBaseComponent.from_relcat_config(config, save_dir_path)
    device = torch.device(
        "cuda" if torch.cuda.is_available() and
        component.relcat_config.general.device != "cpu" else "cpu")

    rel_cat = RelCAT(
        base_tokenizer, cdb=cdb,
        config=component.relcat_config, task=component.task)
    rel_cat.device = device
    rel_cat.component = component
    return rel_cat


def load_cnf(cnf_path: str) -> ConfigRelCAT:
    with open(cnf_path) as f1:
        data = json.load(f1)
    data = fix_old_style_cnf(data)
    cnf = ConfigRelCAT.model_validate(data)
    cnf.comp_name = RelCATAddon.addon_type
    return cnf


def get_rel_cat_from_old(cdb: CDB, old_path: str, tokenizer: BaseTokenizer
                         ) -> RelCATAddon:
    """Convert a v1 RelCAT folder to a v2 RelCAT.

    Args:
        cdb (CDB): The base CDB.
        old_path (str): The v1 RelCAT file path.
        tokenizer (BaseTokenizer): The tokenizer.

    Returns:
        RelCATAddon: The v2 RelCAT.
    """
    cnf = load_cnf(os.path.join(old_path, "config.json"))
    rel_cat = _load_legacy(cdb, tokenizer, cnf, old_path)
    addon = RelCATAddon.create_new(cnf, tokenizer, rel_cat.cdb)
    addon._rel_cat = rel_cat
    return addon


if __name__ == "__main__":
    import sys
    from medcat.config import Config
    cdb = CDB(Config())
    cdb.config.general.nlp.provider = 'spacy'
    rc = get_rel_cat_from_old(
        cdb, sys.argv[1], create_tokenizer("spacy", cdb.config))



import torch
from torch import Tensor, LongTensor
from torch.nn.utils.rnn import pad_sequence


class Pad_Sequence():

    def __init__(self, seq_pad_value: int, label_pad_value: int = -1):
        """Used in rel_cat.py in RelCAT to create DataLoaders for
        train/test datasets.
        collate_fn for dataloader to collate sequences of different input_ids,
        ent1/ent2, and label lengths into a fixed length batch.
        This is applied per batch and not on the whole DataLoader data,
        padded x sequence, y sequence, x lengths and y lengths of batch.

        Args:
            seq_pad_value (int): pad value for input_ids.
            label_pad_value (int): pad value for labels. Defaults to -1.
        """
        self.seq_pad_value: int = seq_pad_value
        self.label_pad_value: int = label_pad_value

    def __call__(self, batch: list[torch.Tensor]) -> tuple[
            Tensor, list, Tensor, LongTensor, LongTensor]:
        """ Pads a batch of input_ids.

        Args:
            batch (list[torch.Tensor]): gets the batch of Tensors from
                RelData.dataset (check __getitem__() method for data returned)
                and pads the token sequence + labels as needed
                See https://pytorch.org/docs/stable/_modules/torch/nn/utils/rnn.html#pad_sequence
                for extra info.

        Returns:
            tuple[Tensor, Tensor, Tensor, LongTensor, LongTensor]: padded data
                padded input ids, ent1&ent2 start token pos, padded labels,
                padded input_id_lengths, padded labels length
        """  # noqa
        sorted_batch = sorted(batch, key=lambda x: x[0].shape[0], reverse=True)

        # input ids
        seqs = [x[0] for x in sorted_batch]
        seqs_padded = pad_sequence(
            seqs, batch_first=True, padding_value=self.seq_pad_value)
        x_lengths = torch.LongTensor([len(x) for x in seqs])

        # label_ids
        labels = list(map(lambda x: x[2], sorted_batch))
        labels_padded = pad_sequence(
            labels, batch_first=True, padding_value=self.label_pad_value)
        y_lengths = torch.LongTensor([len(x) for x in labels])

        ent1_ent2_start_pos = list(map(lambda x: x[1], sorted_batch))

        return (seqs_padded, ent1_ent2_start_pos,
                labels_padded, x_lengths, y_lengths)



from typing import Any
from typing_extensions import TypedDict


CDBInfo = TypedDict(
    "CDBInfo", {
        "Number of concepts": int,
        "Number of names": int,
        "Number of concepts that received training": int,
        "Number of seen training examples in total": int,
        "Average training examples per concept": float,
        "Unsupervised training history": list[dict[str, Any]],
        "Supervised training history": list[dict[str, Any]],
    }
)


ModelCard = TypedDict(
    "ModelCard", {
        "Model ID": str,
        "Last Modified On": str,
        "History (from least to most recent)": list[str],
        'Description': str,
        'Source Ontology': list[str],
        'Location': str,
        'MetaCAT models': list[str],
        'Basic CDB Stats': CDBInfo,
        'Performance': dict[str, Any],
        ('Important Parameters '
         '(Partial view, all available in cat.config)'): dict[str, Any],
        'MedCAT Version': str,
    }
)



import argparse
import json
from pathlib import Path
import logging

from typing import Optional

from medcat.cat import CAT
from medcat.utils.regression.checking import RegressionSuite, TranslationLayer
from medcat.utils.regression.results import (
    Strictness, Finding, STRICTNESS_MATRIX)

logger = logging.getLogger(__name__)


DEFAULT_TEST_SUITE_PATH = Path('configs', 'default_regression_tests.yml')


def show_description():
    logger.info('The various findings and their descriptions:')
    logger.info('')
    logger.info('Class description:')
    logger.info('')
    logger.info(Finding.__doc__.replace("\n    ", "\n"))
    logger.info('')
    for f in Finding:
        logger.info('%s :', f.name)
        logger.info(f.__doc__.replace("\n    ", "\n"))
        logger.info('')
    logger.info('The strictnesses we have available:')
    logger.info('')
    for strictness in Strictness:
        allows = [s.name for s in STRICTNESS_MATRIX[strictness]]
        logger.info('%s: allows %s', strictness.name, allows)
        logger.info('')
    logger.info(
        'NOTE: When using --example-strictness, anything described above '
        'will be omitted from examples (since the are considered correct)')


def main(model_pack_dir: Path, test_suite_file: Path,
         phrases: bool = False, hide_empty: bool = False,
         examples_strictness_str: str = 'STRICTEST',
         jsonpath: Optional[Path] = None, overwrite: bool = False,
         jsonindent: Optional[int] = None,
         strictness_str: str = 'NORMAL',
         max_phrase_length: int = 80,
         use_mct_export: bool = False,
         mct_export_yaml_path: Optional[str] = None,
         only_mct_export_conversion: bool = False,
         only_describe: bool = False,
         require_fully_correct: bool = False,
         edit_distance: tuple[int, int, int] = (0, 0, 0)) -> None:
    """Check test suite against the specifeid model pack.

    Args:
        model_pack_dir (Path): The path to the model pack
        test_suite_file (Path): The path to the test suite YAML
        phrases (bool): Whether to show per-phrase information in a report
        hide_empty (bool): Whether to hide empty cases in a report
        examples_strictness_str (str): The example strictness string.
            Defaults to STRICTEST.
            NOTE: If you set this to 'None', examples will be omitted.
        jsonpath (Optional[Path]): The json path to save the report to
            (if specified)
        overwrite (bool): Whether to overwrite the file if it exists.
            Defaults to False
        jsonindent (int): The indentation for json objects. Defaults to 0
        strictness_str (str): The strictness name. Defaults to NORMAL.
        max_phrase_length (int): The maximum phrase length in examples.
            Defaults to 80.
        use_mct_export (bool): Whether to use a MedCATtrainer export as input.
            Defaults to False.
        mct_export_yaml_path (str): The (optional) path the converted
            MCT export should be saved as YAML at. If not set (or None),
            the MCT export is not saved in YAML format. Defaults to None.
        only_mct_export_conversion (bool): Whether to only deal with the
            MCT export conversion. I.e exit when MCT export conversion is
            done. Defaults to False.
        only_describe (bool): Whether to only describe the finding options
            and exit. Defaults to False.
        require_fully_correct (bool): Whether all cases are required to
            be correct. If set to True, an exit-status of 1 is returned
            unless all (sub)cases are correct. Defaults to False.
        edit_distance (tuple[int, int, int]): The edit distance, the random
            seed, and the number of edited names to pick for each of the
            names. If set to non-0, the specified number of splits, deletes,
            transposes, replaces, or inserts are done to the each name. This
            can be useful for looking at the capability of identifying typos
            in text. However, this can make hte process a lot slower as a
            result. Defaults to (0, 0, 0).

    Raises:
        ValueError: If unable to overwrite file or folder does not exist.
    """
    if only_describe:
        show_description()
        return
    if jsonpath and jsonpath.exists() and not overwrite:
        # check before doing anything so as to not waste time on the tests
        raise ValueError(
            f'Unable to write to existing file {str(jsonpath)} pass '
            '--overwrite to overwrite the file')
    if jsonpath and not jsonpath.parent.exists():
        raise ValueError(
            'Need to specify a file in an existing directory, folder not '
            f'found: {str(jsonpath)}')
    logger.info('Loading RegressionChecker from yaml: %s', test_suite_file)
    if not use_mct_export:
        rc = RegressionSuite.from_yaml(str(test_suite_file))
    else:
        rc = RegressionSuite.from_mct_export(str(test_suite_file))
        if mct_export_yaml_path:
            logger.info('Writing MCT export in YAML to %s',
                        str(mct_export_yaml_path))
            with open(mct_export_yaml_path, 'w') as f:
                f.write(rc.to_yaml())
            if only_mct_export_conversion:
                logger.info("Done with conversion - exiting")
                return
    logger.info('Loading model pack from file: %s', model_pack_dir)
    cat: CAT = CAT.load_model_pack(str(model_pack_dir))
    logger.info('Checking the current status')
    res = rc.check_model(cat, TranslationLayer.from_CDB(cat.cdb),
                         edit_distance=edit_distance,
                         use_diacritics=cat.config.general.diacritics)
    cat.config.general
    strictness = Strictness[strictness_str]
    if examples_strictness_str in ("None", "N/A"):
        examples_strictness = None
    else:
        examples_strictness = Strictness[examples_strictness_str]
    if jsonpath:
        logger.info('Writing to %s', str(jsonpath))
        dumped = res.model_dump(strictness=examples_strictness)
        jsonpath.write_text(json.dumps(dumped, indent=jsonindent))
    else:
        logger.info(res.get_report(phrases_separately=phrases,
                    hide_empty=hide_empty,
                    examples_strictness=examples_strictness,
                    strictness=strictness, phrase_max_len=max_phrase_length))
    if require_fully_correct:
        total, success = res.calculate_report(
            phrases_separately=phrases, hide_empty=hide_empty,
            examples_strictness=examples_strictness, strictness=strictness,
            phrase_max_len=max_phrase_length)[:2]
        if total != success:
            exit(1)


def tuple3_parser(arg: str) -> tuple[int, int, int]:
    parts = arg.strip("()").split(',')
    if len(parts) != 3:
        raise argparse.ArgumentTypeError("Tuple must be in the form (x, y, z)")
    try:
        return (int(parts[0]), int(parts[1]), int(parts[2]))
    except ValueError:
        raise argparse.ArgumentTypeError("Tuple must be in the form (x, y, z)")


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'modelpack', help='The model pack against which to check', type=Path)
    parser.add_argument(
        'test_suite',
        help='YAML formatted file containing the regression test suite'
             'The default value (and example) is at '
             f'`{DEFAULT_TEST_SUITE_PATH}`',
        default=DEFAULT_TEST_SUITE_PATH, nargs='?', type=Path)
    parser.add_argument(
        '--silent', '-s',
        help='Make the operation silent (i.e ignore console output)',
        action='store_true')
    parser.add_argument(
        '--verbose', '-debug', help='Enable debug/verbose mode',
        action='store_true')
    parser.add_argument(
        '--phrases', '-p', help='Include per-phrase information in report',
        action='store_true')
    parser.add_argument(
        '--noempty', help='Hide empty cases in report', action='store_true')
    parser.add_argument(
        '--example-strictness',
        help='The strictness of examples. Set to None to disable. '
             'This defaults to STRICTEST to show all non-identical examples. ',
        choices=[strictness.name for strictness in Strictness] + ["None"],
        default=Strictness.STRICTEST.name)
    parser.add_argument(
        '--jsonfile', help='Save report to a json file', type=Path)
    parser.add_argument(
        '--overwrite', help='Whether to overwrite save file',
        action='store_true')
    parser.add_argument(
        '--jsonindent', help='The json indent', type=int, default=None)
    parser.add_argument(
        '--strictness', help='The strictness to consider success.',
        choices=[strictness.name for strictness in Strictness],
        default=Strictness.NORMAL.name)
    parser.add_argument(
        '--max-phrase-length', help='The maximum phrase length in examples.',
        type=int, default=80)
    parser.add_argument(
        '--from-mct-export', help='Whether to load the regression suite from '
        'a MedCATtrainer export (.json) instead of a YAML format (default).',
        action='store_true')
    parser.add_argument(
        '--mct-export-yaml', help='The YAML file path to safe a convert MCT '
        'export as. Only useful alongside `--from-mct-export` option and an '
        'MCT export passed as the test suite.',
        type=str, default=None)
    parser.add_argument(
        '--only-conversion', help='Whether to load only deal with the '
        'MCT export conversion. Only useful alongside `--from-mct-export` '
        'and `--mct-export-yaml`', action='store_true')
    parser.add_argument(
        '--only-describe', help='Only describe the various findings and exit.',
        action='store_true')
    parser.add_argument(
        '--require-fully-correct', help='Require the regression test to be '
        'fully correct. If set, a non-zero exit status is returned unless all '
        'cases are successful (100%%). This can be useful for (e.g) CI '
        'workflow integration.',
        action='store_true')
    parser.add_argument(
        '--edit-distance', help='Set the edit distance of each of the names. '
        'If set, each name tested will have the specified number of '
        'characters changed. This can be useful to determine the versatility '
        'of the model in terms of recognising typos. Defauts to 0 '
        '(i.e no change). You need to provide 3 numbers in the format '
        '`(N, R, P)` where `N` is the edit distance, `R` is the random seed, '
        'and `P` is the number of choices to make.',
        type=tuple3_parser, default=(0, 0, 0))
    args = parser.parse_args()
    if not args.silent:
        logger.addHandler(logging.StreamHandler())
        logger.setLevel('INFO')
    if args.verbose:
        from medcat.utils.regression import logger as regr_logger
        regr_logger.setLevel(logging.DEBUG)
        regr_logger.addHandler(logging.StreamHandler())
    main(args.modelpack, args.test_suite,
         phrases=args.phrases,
         hide_empty=args.noempty,
         examples_strictness_str=args.example_strictness,
         jsonpath=args.jsonfile,
         overwrite=args.overwrite,
         jsonindent=args.jsonindent,
         strictness_str=args.strictness,
         max_phrase_length=args.max_phrase_length,
         use_mct_export=args.from_mct_export,
         mct_export_yaml_path=args.mct_export_yaml,
         only_mct_export_conversion=args.only_conversion,
         only_describe=args.only_describe,
         require_fully_correct=args.require_fully_correct,
         edit_distance=args.edit_distance)



from enum import Enum, auto
from typing import Optional, Iterable, cast
import json
import pydantic

from medcat.utils.regression.targeting import TranslationLayer, FinalTarget
from medcat.utils.regression.utils import (
    limit_str_len, add_doc_strings_to_enum)
from medcat.data.entities import Entity


class Finding(Enum):
    """Describes whether or how the finding verified.

    The idea is that we know where we expect the entity to be recognised
    and the enum constants describe how the recognition compared to the
    expectation.

    In essence, we want to know the relative positions of the two pairs of
    numbers (character numbers):
    - Expected Start, Expected End
    - Recognised Start, Recognised End

    We can model this as 4 numbers on the number line. And we want to know
    their position relative to each other.
    For example, if the expected positions are marked with * and recognised
    positions with #, we may have something like:
    ___*__#_______#*______________
    Which would indicate that there is a partial, but smaller span recognised.
    """
    # same CUIs
    IDENTICAL = auto()
    """The CUI and the span recognised are identical to what was expected."""
    BIGGER_SPAN_RIGHT = auto()
    """The CUI is the same, but the recognised span is longer on the right.

    If we use the notation from the class doc string, e.g:
    _*#__*__#"""
    BIGGER_SPAN_LEFT = auto()
    """The CUI is the same, but the recognised span is longer on the left.

    If we use the notation from the class doc string, e.g:
    _#_*__*#_"""
    BIGGER_SPAN_BOTH = auto()
    """The CUI is the same, but the recognised span is longer on both sides.

    If we use the notation from the class doc string, e.g:
    _#__*__*__#_"""
    SMALLER_SPAN = auto()
    """The CUI is the same, but the recognised span is smaller.

    If we use the notation from the class doc string, e.g:
    _*_#_#_*_ (neither start nor end match)
    _*#_#_*__ (start matches, but end is before expected)
    _*__#_#*_ (end matches, but start is after expected)"""
    PARTIAL_OVERLAP = auto()
    """The CUI is the same, but the span overlaps partially.

    If we use the notation from the class doc string, e.g:
    _*_#__*_#_ (starts between expected start and end, but ends beyond)
    _#_*_#_*__ (start before expected start, but ends between expected start and end)""" # noqa
    # slightly different CUIs
    FOUND_DIR_PARENT = auto()
    """The recognised CUI is a parent of the expected CUI but the span is an exact match.""" # noqa
    FOUND_DIR_GRANDPARENT = auto()
    """The recognised CUI is a grandparent of the expected CUI but the span is an exact match.""" # noqa
    FOUND_ANY_CHILD = auto()
    """The recognised CUI is a child of the expected CUI but the span is an exact match.""" # noqa
    FOUND_CHILD_PARTIAL = auto()
    """The recognised CUI is a child yet the match is only partial (smaller/bigger/partial).""" # noqa
    FOUND_OTHER = auto()
    """Found another CUI in the same span."""
    FAIL = auto()
    """The concept was not recognised in any meaningful way."""

    def has_correct_cui(self) -> bool:
        """Whether the finding found the correct concept.

        Returns:
            bool: Whether the correct concept was found.
        """
        return self in (
            Finding.IDENTICAL, Finding.BIGGER_SPAN_RIGHT,
            Finding.BIGGER_SPAN_LEFT,
            Finding.BIGGER_SPAN_BOTH, Finding.SMALLER_SPAN,
            Finding.PARTIAL_OVERLAP
        )

    @classmethod
    def determine(cls, exp_cui: str, exp_start: int, exp_end: int,
                  tl: TranslationLayer,
                  found_entities: dict[int, Entity],
                  strict_only: bool = False,
                  check_children: bool = True, check_parent: bool = True,
                  check_grandparent: bool = True
                  ) -> tuple['Finding', Optional[str]]:
        """Determine the finding type based on the input

        Args:
            exp_cui (str): Expected CUI.
            exp_start (int): Expected span start.
            exp_end (int): Expected span end.
            tl (TranslationLayer): The translation layer.
            found_entities (dict[int, Entity]):
                The entities found by the model.
            strict_only (bool): Whether to use a strict-only mode
                (either identical or fail). Defaults to False.
            check_children (bool): Whether to check the children.
                Defaults to True.
            check_parent (bool): Whether to check for parent(s).
                Defaults to True.
            check_grandparent (bool): Whether to check for grandparent(s).
                Defaults to True.

        Returns:
            tuple['Finding', Optional[str]]:
                The type of finding determined, and the alternative.
        """
        return FindingDeterminer(exp_cui, exp_start, exp_end,
                                 tl, found_entities, strict_only,
                                 check_children, check_parent,
                                 check_grandparent).determine()


# NOTE: add doc strings to enum constants
add_doc_strings_to_enum(Finding)


class FindingDeterminer:
    """A helper class to determine the type of finding.

    This is mostly useful to split the responsibilities of
    looking at children/parents as well as to keep track of
    the already-checked children to avoid infinite recursion
    (which could happen in - e.g - a SNOMED model).

    Args:
        exp_cui (str): The expected CUI.
        exp_start (int): The expected span start.
        exp_end (int): The expected span end.
        tl (TranslationLayer): The translation layer.
        found_entities (dict[str, Entity]):
            The entities found by the model.
        strict_only (bool):
            Whether to use strict-only mode (either identical or fail).
            Defaults to False.
        check_children (bool):
            Whether or not to check the children. Defaults to True.
        check_parent (bool):
            Whether to check for parent(s). Defaults to True.
        check_grandparent (bool):
            Whether to check for granparent(s). Defaults to True.
    """

    def __init__(self, exp_cui: str,
                 exp_start: int,
                 exp_end: int,
                 tl: TranslationLayer,
                 found_entities: dict[int, Entity],
                 strict_only: bool = False,
                 check_children: bool = True,
                 check_parent: bool = True,
                 check_grandparent: bool = True,) -> None:
        self.exp_cui = exp_cui
        self.exp_start = exp_start
        self.exp_end = exp_end
        self.tl = tl
        self.found_entities = found_entities
        self.strict_only = strict_only
        self.check_children = check_children
        self.check_parent = check_parent
        self.check_grandparent = check_grandparent
        # helper for children to avoid infinite recursion
        self._checked_children: set[str] = set()

    def _determine_raw(self, start: int, end: int) -> Optional[Finding]:
        """Determines the raw SPAN-ONLY finding.

        I.e this assumes the concept is appropriate.
        It will return None if there is no overlapping span.

        Args:
            start (int): The start of the span.
            end (int): The end of the span.

        Raises:
            MalformedFinding:
                If the start is greater than the end.
            MalformedFinding:
                If the expected start is greater than the expected end.

        Returns:
            Optional[Finding]: The finding, if a match is found.
        """
        if end < start:
            raise MalformedFinding(
                f"The end ({end}) is smaller than the start ({start})")
        elif self.exp_end < self.exp_start:
            raise MalformedFinding(
                f"The expected end ({self.exp_end}) is "
                f"smaller than the expected start ({self.exp_start})")
        if self.strict_only:
            if start == self.exp_start and end == self.exp_end:
                return Finding.IDENTICAL
            return None
        if start < self.exp_start:
            if end < self.exp_start:
                return None
            elif end < self.exp_end:
                return Finding.PARTIAL_OVERLAP  # TODO - distinguish[overlap]?
            elif end == self.exp_end:
                return Finding.BIGGER_SPAN_LEFT
            return Finding.BIGGER_SPAN_BOTH
        elif start == self.exp_start:
            if end < self.exp_end:
                return Finding.SMALLER_SPAN  # TODO - distinguish[smaller]?
            elif end == self.exp_end:
                return Finding.IDENTICAL
            return Finding.BIGGER_SPAN_RIGHT
        elif start > self.exp_start and start <= self.exp_end:
            if end < self.exp_end:
                return Finding.SMALLER_SPAN  # TODO - distinguish[smaller]?
            elif end == self.exp_end:
                return Finding.SMALLER_SPAN  # TODO - distinguish[smaller]?
            return Finding.PARTIAL_OVERLAP   # TODO - distinguish[overlap]?
        # if start > exp_end -> no match
        return None

    def _get_strict(self) -> Optional[Finding]:
        if not self.found_entities:
            return Finding.FAIL
        for entity in self.found_entities.values():
            start, end, cui = entity['start'], entity['end'], entity['cui']
            if cui == self.exp_cui:
                raw_find = self._determine_raw(start, end)
                if raw_find:
                    return raw_find
        if self.strict_only:
            return Finding.FAIL
        return None

    def _check_parents(self) -> Optional[tuple[Finding, Optional[str]]]:
        parents = self.tl.get_direct_parents(self.exp_cui)
        for parent in parents:
            finding, wcui = Finding.determine(
                parent, self.exp_start, self.exp_end,
                self.tl,
                self.found_entities,
                check_children=False,
                check_parent=self.check_grandparent,
                check_grandparent=False)
            if finding is Finding.IDENTICAL:
                return Finding.FOUND_DIR_PARENT, parent
            if finding is Finding.FOUND_DIR_PARENT:
                return Finding.FOUND_DIR_GRANDPARENT, wcui
        return None

    def _check_children(self) -> Optional[tuple[Finding, Optional[str]]]:
        children = self.tl.get_direct_children(self.exp_cui)
        for child in children:
            finding, wcui = Finding.determine(
                child, self.exp_start, self.exp_end,
                self.tl,
                self.found_entities,
                check_children=True,
                check_parent=False,
                check_grandparent=False)
            if finding in (Finding.IDENTICAL, Finding.FOUND_ANY_CHILD):
                alt_cui = child if finding == Finding.IDENTICAL else wcui
                return Finding.FOUND_ANY_CHILD, alt_cui
            elif finding.has_correct_cui():
                # i.e a partial match with same CUI
                return Finding.FOUND_CHILD_PARTIAL, child
            elif finding is Finding.FOUND_CHILD_PARTIAL:
                return finding, wcui
            self._checked_children.add(child)
        return None

    def _descr_cui(self, cui: Optional[str]) -> Optional[str]:
        if cui is None:
            return None
        return f"{cui} ({self.tl.get_preferred_name(cui)})"

    def _find_diff_cui(self) -> Optional[tuple[Finding, str]]:
        for entity in self.found_entities.values():
            start, end, cui = entity['start'], entity['end'], entity['cui']
            if start == self.exp_start and end == self.exp_end:
                return Finding.FOUND_OTHER, cui
        return None

    def determine(self) -> tuple[Finding, Optional[str]]:
        """Determine the finding based on the given information.

        First, the strict check is done (either identical or not).
        Then, parents are checked (if required).
        After that, children are checked (if required).

        Returns:
            tuple[Finding, Optional[str]]: The appropriate finding,
                and the alternative (if applicable).
        """
        finding, cui = self._determine()
        # NOTE: the point of this wrapper method is to add the preferred name
        #       to the CUI in one place and one place only
        return finding, self._descr_cui(cui)

    def _determine(self) -> tuple[Finding, Optional[str]]:
        finding = self._get_strict()
        if finding is not None:
            return finding, None
        if self.check_parent:
            fpar = self._check_parents()
            if fpar is not None:
                return fpar
        if self.check_children:
            self._checked_children.add(self.exp_cui)
            fch = self._check_children()
            if fch is not None:
                return fch
        fdcui = self._find_diff_cui()
        return fdcui or (Finding.FAIL, None)


class Strictness(Enum):
    """The total strictness on which to judge the results."""
    STRICTEST = auto()
    """The strictest option which only allows identical findings."""
    STRICT = auto()
    """A strict option which allows identical or children."""
    NORMAL = auto()
    """Normal strictness also allows partial overlaps on target concept and children.""" # noqa
    LENIENT = auto()
    """Lenient stictness also allows parents and grandparents."""
    ANYTHING = auto()
    """Anything stricness allows ANY finding.

    This would generally only be relevant when disabling examples
    for results descriptors."""


STRICTNESS_MATRIX: dict[Strictness, set[Finding]] = {
    Strictness.STRICTEST: {Finding.IDENTICAL},
    Strictness.STRICT: {Finding.IDENTICAL, Finding.FOUND_ANY_CHILD},
    Strictness.NORMAL: {
        Finding.IDENTICAL, Finding.FOUND_ANY_CHILD,
        Finding.FOUND_CHILD_PARTIAL,
        Finding.BIGGER_SPAN_RIGHT, Finding.BIGGER_SPAN_LEFT,
        Finding.BIGGER_SPAN_BOTH,
        Finding.SMALLER_SPAN, Finding.PARTIAL_OVERLAP
    },
    Strictness.LENIENT: {
        Finding.IDENTICAL, Finding.FOUND_ANY_CHILD,
        Finding.BIGGER_SPAN_RIGHT, Finding.BIGGER_SPAN_LEFT,
        Finding.BIGGER_SPAN_BOTH,
        Finding.SMALLER_SPAN, Finding.PARTIAL_OVERLAP,
        Finding.FOUND_DIR_PARENT, Finding.FOUND_DIR_GRANDPARENT,
    },
    Strictness.ANYTHING: set(Finding),
}


class SingleResultDescriptor(pydantic.BaseModel):
    """The result descriptor.

    This class is responsible for keeping track of all the
    findings (i.e how many were found to be identical) as
    well as the examples of the finding on a per-target
    basis for further analysis.
    """
    name: str
    """The name of the part that was checked"""
    findings: dict[Finding, int] = {}
    """The description of failures"""
    examples: list[tuple[FinalTarget, tuple[Finding, Optional[str]]]] = []
    """The examples of non-perfect alignment."""

    def report_success(self, target: FinalTarget,
                       found: tuple[Finding, Optional[str]]) -> None:
        """Report a test case and its successfulness.

        Args:
            target (FinalTarget): The target configuration
            found (tuple[Finding, Optional[str]]):
                Whether or not the check was successful
        """
        finding, _ = found
        if finding not in self.findings:
            self.findings[finding] = 0
        self.findings[finding] += 1
        self.examples.append((target, found))

    def get_report(self) -> str:
        """Get the report associated with this descriptor

        Returns:
            str: The report string
        """
        total = sum(self.findings.values())
        ret_vals = [f"Tested '{self.name}' for a total of {total} cases:"]
        ret_vals.extend([
            f"{f.name:24s}:{self.findings[f]:10d} "
            f"({100 * self.findings[f] / total if total > 0 else 0:5.2f}%)"
            # NOTE iterating over Finding so the order is the same
            #      as in the enum
            for f in Finding if f in self.findings
        ])
        return "\n".join(ret_vals)

    def model_dump(self, **kwargs) -> dict:
        if 'strictness' in kwargs:
            kwargs = kwargs.copy()  # so if used elsewhere, keeps the kwarg
            strict_raw = kwargs.pop('strictness')
            if isinstance(strict_raw, Strictness):
                strictness = strict_raw
            elif isinstance(strict_raw, str):
                strictness = Strictness[strict_raw]
            else:
                raise ValueError(f"Unknown stircntess specified: {strict_raw}")
        else:
            strictness = Strictness.NORMAL
        # avoid serialising multiple times
        if 'exclude' in kwargs and kwargs['exclude'] is not None:
            exclude: set = kwargs['exclude']
        else:
            exclude = set()
            kwargs['exclude'] = exclude
        exclude.update(('findings', 'examples'))
        serialized_dict = {
            key.name: value for key, value in self.findings.items()
        }
        serialized_examples = [
            (ft.model_dump(**kwargs), (f[0].name, f[1]))
            for ft, f in self.examples
            # only count if NOT in strictness matrix (i.e 'failures')
            if f[0] not in STRICTNESS_MATRIX[strictness]
        ]
        model_dict = cast(pydantic.BaseModel, super()).model_dump(**kwargs)
        model_dict['findings'] = serialized_dict
        model_dict['examples'] = serialized_examples
        return model_dict

    def json(self, **kwargs) -> str:
        d = self.model_dump(**kwargs)
        return json.dumps(d)


class ResultDescriptor(SingleResultDescriptor):
    """The overarching result descriptor that handles multiple phrases.

    This class keeps track of the results on a per-phrase basis and
    can be used to get the overall report and/or iterate over examples.
    """
    per_phrase_results: dict[str, SingleResultDescriptor] = {}

    def report(self, target: FinalTarget,
               finding: tuple[Finding, Optional[str]]) -> None:
        """Report a test case and its successfulness

        Args:
            target (FinalTarget): The final targe configuration
            finding (tuple[Finding, Optional[str]]):
                To what extent the concept was recognised
        """
        phrase = target.final_phrase
        super().report_success(target, finding)
        if phrase not in self.per_phrase_results:
            self.per_phrase_results[phrase] = SingleResultDescriptor(
                name=phrase)
        self.per_phrase_results[phrase].report_success(target, finding)

    def iter_examples(self, strictness_threshold: Strictness
                      ) -> Iterable[tuple[FinalTarget,
                                          tuple[Finding, Optional[str]]]]:
        """Iterate suitable examples.

        The strictness threshold at which to include examples.

        Any finding that is assumed to be "correct enough" according to
        the strictness matrix for this threshold will be withheld from
        examples.

        In simpler terms, if the finding is NOT in the strictness matrix
        for this strictness, the example is recorded.

        NOTE: To disable example keeping, set the threshold to
              Strictness.ANYTHING.

        Args:
            strictness_threshold (Strictness): The strictness threshold.

        Yields:
            Iterable[tuple[FinalTarget, tuple[Finding, Optional[str]]]]:
                The placeholder, phrase, finding, CUI, and name.
        """
        phrases = sorted(self.per_phrase_results.keys())
        for phrase in phrases:
            srd = self.per_phrase_results[phrase]
            # sort by finding 1st, found CUI 2nd, and used name 3rd
            sorted_examples = sorted(
                srd.examples, key=lambda tf: (
                    tf[1][0].name, str(tf[1][1]), tf[0].name))
            for target, finding in sorted_examples:
                if finding[0] not in STRICTNESS_MATRIX[strictness_threshold]:
                    yield target, finding

    def get_report(self, phrases_separately: bool = False) -> str:
        """Get the report associated with this descriptor

        Args:
            phrases_separately (bool): Whether to output descriptor
                for each phrase separately

        Returns:
            str: The report string
        """
        sr = super().get_report()
        if not phrases_separately:
            return sr
        children = '\n'.join([srd.get_report()
                             for srd in self.per_phrase_results.values()])
        return sr + '\n\t\t' + children.replace('\n', '\n\t\t')

    def model_dump(self, **kwargs) -> dict:
        if 'exclude' in kwargs and kwargs['exclude'] is not None:
            exclude: set = kwargs['exclude']
        else:
            exclude = set()
            kwargs['exclude'] = exclude
        # NOTE: ignoring here so that examples are only present
        #       in the per phrase part
        exclude.update(('examples', 'per_phrase_results'))
        d = cast(pydantic.BaseModel, super()).model_dump(**kwargs)
        if 'examples' in d:
            # NOTE: I don't really know why, but the examples still
            #       seem to be a part of the resulting dict, so I need
            #       to explicitly remove them
            del d['examples']
        # NOTE: need to propagate here manually so the strictness keyword
        #       makes sense and doesn't cause issues due being to
        #       unexpected keyword
        per_phrase_results = {
            phrase: res.model_dump(**kwargs) for phrase, res in
            sorted(self.per_phrase_results.items(), key=lambda it: it[0])
        }
        d['per_phrase_results'] = per_phrase_results
        return d


class MultiDescriptor(pydantic.BaseModel):
    """The descriptor of results over multiple different results (parts).

    The idea is that this would likely be used with a regression suite
    and it would incorporate all the different regression cases it describes.
    """
    name: str
    """The name of the collection being checked"""
    parts: list[ResultDescriptor] = []
    """The parts kept track of"""

    @property
    def findings(self) -> dict[Finding, int]:
        """The total findings.

        Returns:
            dict[Finding, int]: The total number of successes.
        """
        totals: dict[Finding, int] = {}
        for part in self.parts:
            for f, val in part.findings.items():
                if f not in totals:
                    totals[f] = val
                else:
                    totals[f] += val
        return totals

    def iter_examples(self, strictness_threshold: Strictness
                      ) -> Iterable[
                          tuple[FinalTarget, tuple[Finding, Optional[str]]]]:
        """Iterate over all relevant examples.

        Only examples that are not in the strictness matrix for the specified
        threshold will be used.

        Args:
            strictness_threshold (Strictness): The threshold of avoidance.

        Yields:
            Iterable[tuple[FinalTarget, tuple[Finding, Optional[str]]]]:
                The examples
        """
        for descr in self.parts:
            yield from descr.iter_examples(
                strictness_threshold=strictness_threshold)

    def _get_part_report(self, part: ResultDescriptor,
                         allowed_findings: set[Finding],
                         total_findings: dict[Finding, int],
                         hide_empty: bool,
                         examples_strictness: Optional[Strictness],
                         phrases_separately: bool,
                         phrase_max_len: int,
                         ) -> tuple[str, int, int, int]:
        if hide_empty and len(part.findings) == 0:
            return '', 0, 0, 0
        total_total, total_s, total_f = 0, 0, 0
        for f, val in part.findings.items():
            if f not in total_findings:
                total_findings[f] = val
            else:
                total_findings[f] += val
            total_total += val
            if f in allowed_findings:
                total_s += val
            else:
                total_f += val
        cur_add = '\t' + \
            part.get_report(phrases_separately=phrases_separately).replace(
                '\n', '\n\t\t')
        if examples_strictness is not None:
            latest_phrase = ''
            for target, found in part.iter_examples(
                    strictness_threshold=examples_strictness):
                finding, ocui = found
                if latest_phrase == '':
                    # add header only if there's failures to include
                    cur_add += (
                        f"\n\t\tExamples at {examples_strictness} strictness")
                if latest_phrase != target.final_phrase:
                    short_phrase = limit_str_len(
                        target.final_phrase, max_length=phrase_max_len,
                        keep_front=phrase_max_len // 2,
                        keep_rear=phrase_max_len // 2 - 10)
                    cur_add += f"\n\t\tWith phrase: {repr(short_phrase)}"
                    latest_phrase = target.final_phrase
                found_cui_descr = f' [{ocui}]' if ocui else ''
                cur_add += (
                    f'\n\t\t\t{finding.name}{found_cui_descr} for '
                    f'placeholder {target.placeholder} '
                    f'with CUI {repr(target.cui)} and name {repr(target.name)}'
                    )
        return cur_add, total_total, total_s, total_f

    def calculate_report(
            self, phrases_separately: bool = False,
            hide_empty: bool = False,
            examples_strictness: Optional[Strictness] = Strictness.STRICTEST,
            strictness: Strictness = Strictness.NORMAL,
            phrase_max_len: int = 80) -> tuple[int, int, int, str, int]:
        """Calculate some of the major parts of the report.

        Args:
            phrases_separately (bool): Whether to include
                per-phrase information
            hide_empty (bool): Whether to hide empty cases
            examples_strictness (Optional[Strictness.STRICTEST]):
                What level of strictness to show for examples.
                Set to None to disable examples.
                Defaults to Strictness.STRICTEST.
            strictness (Strictness): The strictness of the
                success / fail overview. Defaults to Strictness.NORMAL.
            phrase_max_len (int): The maximum length of the phrase in examples.
                Defaults to 80.

        Returns:
            tuple[int, int, int, int, str]: The total number of examples,
                the total successes, the total failures,
                the delegated part, and the number of empty
        """
        del_out = []  # delegation
        total_findings: dict[Finding, int] = {}
        total_s, total_f = 0, 0
        allowed_findings = STRICTNESS_MATRIX[strictness]
        total_total = 0
        nr_of_empty = 0
        for part in self.parts:
            (cur_add, total_total_add,
             total_s_add, total_f_add) = self._get_part_report(
                 part, allowed_findings, total_findings, hide_empty,
                 # NOTE: using STRICTEST strictness for examples means
                 #       that all but IDENTICAL examples will be shown
                 examples_strictness, phrases_separately, phrase_max_len)
            if hide_empty and total_total_add == 0:
                nr_of_empty += 1
            else:
                total_total += total_total_add
                total_s += total_s_add
                total_f += total_f_add
                del_out.append(cur_add)
        delegated = '\n'.join(del_out)
        return total_total, total_s, total_f, delegated, nr_of_empty

    def get_report(
            self, phrases_separately: bool,
            hide_empty: bool = False,
            examples_strictness: Optional[Strictness] = Strictness.STRICTEST,
            strictness: Strictness = Strictness.NORMAL,
            phrase_max_len: int = 80) -> str:
        """Get the report associated with this descriptor

        Args:
            phrases_separately (bool):
                Whether to include per-phrase information
            hide_empty (bool): Whether to hide empty cases
            examples_strictness (Optional[Strictness.STRICTEST]):
                What level of strictness to show for examples.
                Set to None to disable examples.
                Defaults to Strictness.STRICTEST.
            strictness (Strictness): The strictness of the
                success / fail overview.
                Defaults to Strictness.NORMAL.
            phrase_max_len (int): The maximum length of the phrase in examples.
                Defaults to 80.

        Returns:
            str: The report string
        """
        (total_total, total_s, total_f,
         delegated, nr_of_empty) = self.calculate_report(
             phrases_separately=phrases_separately,
             hide_empty=hide_empty,
             examples_strictness=examples_strictness,
             strictness=strictness,
             phrase_max_len=phrase_max_len)
        empty_text = ''
        allowed_findings = STRICTNESS_MATRIX[strictness]
        if hide_empty:
            empty_text = (
                f' A total of {nr_of_empty} cases '
                'did not match any CUIs and/or names.')
        ret_vals = [
            f'A total of {len(self.parts)} parts were kept track of within '
            f'the group "{self.name}".\n'
            f'                    And a total of {total_total} '
            f'(sub)cases were checked.{empty_text}']
        allowed_fingings_str = sorted([f.name for f in allowed_findings])
        ret_vals.extend([
            f"At the strictness level of {strictness} "
            f"(allowing {allowed_fingings_str}):",
            f"The number of total successful (sub) cases: {total_s} "
            f"({100 * total_s / total_total if total_total > 0 else 0:5.2f}%)",
            f"The number of total failing (sub) cases   : {total_f} "
            f"({100 * total_f / total_total if total_total > 0 else 0:5.2f}%)"
        ])
        ret_vals.extend([
            f"{f.name:24s}:{self.findings[f]:10d} "
            f"({ft100 / total_total if total_total > 0 else 0:5.2f}%)"
            # NOTE iterating over Finding so the order is the same
            #      as in the enum
            for f in Finding if f in self.findings
            # NOTE: this is just for the line length, really
            and (ft100 := 100 * self.findings[f]) >= 0
        ])
        return "\n".join(ret_vals) + f"\n{delegated}"

    def model_dump(self, **kwargs) -> dict:
        if 'strictness' in kwargs:
            strict_raw = kwargs.pop('strictness')
            if isinstance(strict_raw, Strictness):
                strictness = strict_raw
            elif isinstance(strict_raw, str):
                strictness = Strictness[strict_raw]
            else:
                raise ValueError(f"Unknown stircntess specified: {strict_raw}")
        else:
            strictness = Strictness.NORMAL
        out_dict = cast(pydantic.BaseModel, super()).model_dump(
            exclude={'parts'}, **kwargs)
        out_dict['parts'] = [part.model_dump(strictness=strictness)
                             for part in self.parts]
        return out_dict


class MalformedFinding(ValueError):

    def __init__(self, *args: object) -> None:
        super().__init__(*args)



from medcat.utils.import_utils import ensure_optional_extras_installed
import medcat

from .meta_cat import (MetaCAT, MetaCATAddon,
                       get_meta_annotations, MetaAnnotationValue)


__all__ = ["MetaCAT", "MetaCATAddon",
           "get_meta_annotations", "MetaAnnotationValue"]

# NOTE: the _ is converted to - automatically
_EXTRA_NAME = "meta-cat"


ensure_optional_extras_installed(medcat.__name__, _EXTRA_NAME)



from typing import Callable, Dict, Optional
import json

from medcat.data.mctexport import count_anns_per_concept
from medcat.cdb import CDB
from medcat.cdb.concepts import get_new_cui_info
from medcat.config.config import Config


def replace_entities_in_text(text: str,
                             entities: Dict,
                             get_cui_name: Callable[[str], str],
                             redact: bool = False) -> str:
    new_text = str(text)
    for ent in sorted(entities.values(), key=lambda ent: ent['start'],
                      reverse=True):
        r = "*" * (ent['end'] - ent['start']
                 ) if redact else get_cui_name(ent['cui'])
        new_text = new_text[:ent['start']] + f'[{r}]' + new_text[ent['end']:]
    return new_text


def make_or_update_cdb(json_path: str, cdb: Optional[CDB] = None,
                       min_count: int = 0) -> CDB:
    """Creates a new CDB or updates an existing one with new
    concepts if the cdb argument is provided. All concepts that are less
    frequent than min_count will be ignored.

    Args:
        json_path (str): The json path
        cdb (Optional[CDB]): The CDB if present. Defaults to None.
        min_count (int): Minimum count to include. Defaults to 0.

    Returns:
        CDB: The same or new CDB.
    """
    with open(json_path) as f:
        json_data = json.load(f)
    cui2cnt = count_anns_per_concept(json_data)
    if cdb is None:
        cdb = CDB(config=Config())

    for cui in cui2cnt.keys():
        if cui2cnt[cui] > min_count:
            # We are adding only what is needed
            cinfo = cdb.cui2info.get(cui, None)
            if cinfo is None:
                cinfo = get_new_cui_info(
                    cui=cui, preferred_name=cui, names={cui})
                cdb.cui2info[cui] = cinfo
            else:
                cinfo['names'] = set([cui])
                cinfo['preferred_name'] = cui
    return cdb



import os
from typing import Optional, Union, overload
from tokenizers import ByteLevelBPETokenizer

from medcat.components.addons.meta_cat.mctokenizers.tokenizers import (
    TokenizerWrapperBase)


FAKE_TOKENIZER_PATH = "#\n/fake-path-not-exist#/"


class TokenizerWrapperBPE(TokenizerWrapperBase):
    """Wrapper around a huggingface tokenizer so that it works with the
    MetaCAT models.

    Args:
        tokenizers.ByteLevelBPETokenizer:
            A huggingface BBPE tokenizer.
    """
    name = 'bbpe'

    def __init__(self, hf_tokenizers: Optional[ByteLevelBPETokenizer] = None
                 ) -> None:
        super().__init__(hf_tokenizers)

        if self.hf_tokenizers is not None:
            # For whatever reason added tokens do not persist with
            # this tokenizer, what to do
            self.hf_tokenizers.add_tokens(['<PAD>'])

    @overload
    def __call__(self, text: str) -> dict: ...

    @overload
    def __call__(self, text: list[str]) -> list[dict]: ...

    def __call__(self, text: Union[str, list[str]]) -> Union[dict, list[dict]]:
        """Tokenize some text

        Args:
            text (Union[str, list[str]]):
                Text/texts to be tokenized.

        Returns:
            Union(dict, list[dict]):
                Dictionary/ies containing `offset_mapping`, `input_ids` and
                `tokens` corresponding to the input text/s.

        Raises:
            Exception: If the input is something other than text or a list
                of text.
        """
        self.hf_tokenizers = self.ensure_tokenizer()

        if isinstance(text, str):
            result = self.hf_tokenizers.encode(text)

            return {'offset_mapping': result.offsets,
                    'input_ids': result.ids,
                    'tokens': result.tokens,
                    }
        elif isinstance(text, list):
            results = self.hf_tokenizers.encode_batch(text)
            output = []
            for result in results:
                output.append({
                    'offset_mapping': result.offsets,
                    'input_ids': result.ids,
                    'tokens': result.tokens,
                })

            return output
        else:
            raise Exception(
                "Unsupported input type, supported: text/list, but got: "
                f"{type(text)}")

    def save(self, dir_path: str) -> None:
        self.hf_tokenizers = self.ensure_tokenizer()

        if self.hf_tokenizers is None:
            raise ValueError("The tokenizer is not loaded yet")

        self.hf_tokenizers.save_model(dir_path, prefix=self.name)

    @classmethod
    def load(cls, dir_path: str, model_variant: Optional[str] = '', **kwargs
             ) -> "TokenizerWrapperBPE":
        tokenizer = cls()
        vocab_file = os.path.join(dir_path, f'{tokenizer.name}-vocab.json')
        merges_file = os.path.join(dir_path, f'{tokenizer.name}-merges.txt')
        tokenizer.hf_tokenizers = ByteLevelBPETokenizer.from_file(
            vocab_filename=vocab_file, merges_filename=merges_file, **kwargs)
        # For whatever reason added tokens do not persist with this tokenizer,
        # so we added it at each load
        tokenizer.hf_tokenizers.add_tokens(['<PAD>'])
        return tokenizer

    @classmethod
    def create_new(cls):
        tokenizer = ByteLevelBPETokenizer()
        return cls(tokenizer)

    def get_size(self) -> int:
        self.hf_tokenizers = self.ensure_tokenizer()
        return self.hf_tokenizers.get_vocab_size()

    def token_to_id(self, token: str) -> Union[int, list[int]]:
        self.hf_tokenizers = self.ensure_tokenizer()
        return self.hf_tokenizers.token_to_id(token)

    def get_pad_id(self) -> Union[int, list[int]]:
        pad = self.token_to_id('<PAD>')
        if pad is None:
            raise Exception(
                "No <PAD> token in the vocabulary of the tokenizer, "
                "please add it")
        return pad



from typing import Iterator, Union, Optional, overload, cast, Any
import logging

from spacy.tokens import Token as SpacyToken
from spacy.tokens import Span as SpacySpan
from spacy.tokens import Doc as SpacyDoc

from medcat.tokenizing.tokens import (BaseToken, MutableToken,
                                      BaseEntity, MutableEntity,
                                      BaseDocument,
                                      UnregisteredDataPathException)


logger = logging.getLogger(__name__)


# set extensions as soon as this is loaded
SpacyToken.set_extension('norm', default=None, force=True)
SpacyToken.set_extension('skip', default=False, force=True)
SpacyToken.set_extension('is_punctuation', default=False, force=True)


class Token:

    def __init__(self, delegate: SpacyToken) -> None:
        self._delegate = delegate

    @property
    def is_punctuation(self) -> bool:
        return self._delegate._.is_punctuation

    @is_punctuation.setter
    def is_punctuation(self, new_val: bool) -> None:
        self._delegate._.is_punctuation = new_val

    @property
    def to_skip(self) -> bool:
        return self._delegate._.skip

    @to_skip.setter
    def to_skip(self, new_val: bool) -> None:
        self._delegate._.skip = new_val

    @property
    def norm(self) -> str:
        return self._delegate._.norm or ''

    @norm.setter
    def norm(self, new_val: str) -> None:
        self._delegate._.norm = new_val

    @property
    def base(self) -> BaseToken:
        return cast(BaseToken, self)

    @property
    def text(self) -> str:
        return self._delegate.text

    @property
    def text_versions(self) -> list[str]:
        return [self.norm, self.lower]

    @property
    def lower(self) -> str:
        return self._delegate.lower_

    @property
    def is_stop(self) -> bool:
        return self._delegate.is_stop

    @property
    def is_digit(self) -> bool:
        return self._delegate.is_digit

    @property
    def is_upper(self) -> bool:
        return self._delegate.is_upper

    @property
    def tag(self) -> Optional[str]:
        return self._delegate.tag_

    @property
    def lemma(self) -> str:
        return self._delegate.lemma_

    @property
    def text_with_ws(self) -> str:
        return self._delegate.text_with_ws

    @property
    def char_index(self) -> int:
        return self._delegate.idx

    @property
    def index(self) -> int:
        return self._delegate.i

    def __str__(self):
        return "M2W[T]:" + str(self._delegate)

    def __repr__(self):
        return repr(str(self))

    def __hash__(self) -> int:
        return hash(self._delegate)

    def __eq__(self, value) -> bool:
        if not isinstance(value, Token):
            return False
        return self._delegate == value._delegate


class Entity:
    _addon_extension_paths: set[str] = set()

    def __init__(self, delegate: SpacySpan) -> None:
        self._delegate = delegate
        # defaults
        self.link_candidates: list[str] = []
        self.context_similarity: float = 0.0
        self.confidence: float = 0.0
        self.cui = ''
        self.id = -1  # TODO - what's the default?
        self.detected_name = ''

    @property
    def base(self) -> BaseEntity:
        return cast(BaseEntity, self)

    def set_addon_data(self, path: str, val: Any) -> None:
        if not self._delegate.has_extension(path):
            raise UnregisteredDataPathException(self.__class__, path)
        return setattr(self._delegate._, path, val)

    def has_addon_data(self, path: str) -> bool:
        return bool(self.get_addon_data(path))

    def get_addon_data(self, path: str) -> Any:
        if not self._delegate.has_extension(path):
            raise UnregisteredDataPathException(self.__class__, path)
        return getattr(self._delegate._, path)

    def get_available_addon_paths(self) -> list[str]:
        return [path for path in self._addon_extension_paths
                if self.has_addon_data(path)]

    @classmethod
    def register_addon_path(cls, path: str, def_val: Any = None,
                            force: bool = True) -> None:
        SpacySpan.set_extension(path, default=def_val, force=force)
        cls._addon_extension_paths.add(path)
        # 'id', 'cui', 'id', 'cui', 'id', 'cui', 'id', 'cui', 'id', 'cui'

    @property
    def text(self) -> str:
        return self._delegate.text

    @property
    def label(self) -> int:
        return self._delegate.label

    @property
    def start_index(self) -> int:
        return self._delegate[0].i

    @property
    def end_index(self) -> int:
        return self._delegate[-1].i

    @property
    def start_char_index(self) -> int:
        return self._delegate.start_char

    @property
    def end_char_index(self) -> int:
        return self._delegate.end_char

    def __iter__(self) -> Iterator[MutableToken]:
        for tkn in self._delegate:
            yield Token(tkn)

    def __len__(self) -> int:
        return len(self._delegate)

    def __str__(self):
        return "M2W[T]:" + str(self._delegate)

    def __repr__(self):
        return repr(str(self))


class Document:
    _addon_extension_paths: set[str] = set()

    def __init__(self, delegate: SpacyDoc) -> None:
        self._delegate = delegate
        self.ner_ents: list[MutableEntity] = []
        self.linked_ents: list[MutableEntity] = []

    @property
    def base(self) -> BaseDocument:
        return cast(BaseDocument, self)

    @property
    def text(self) -> str:
        return self._delegate.text

    @overload
    def __getitem__(self, index: int) -> MutableToken:
        pass

    @overload
    def __getitem__(self, index: slice) -> MutableEntity:
        pass

    def __getitem__(self, index: Union[int, slice]
                    ) -> Union[MutableToken, MutableEntity]:
        delegated = self._delegate[index]
        if isinstance(delegated, SpacyToken):
            return Token(delegated)
        return Entity(delegated)

    def __len__(self) -> int:
        return len(self._delegate)

    def get_tokens(self, start_index: int, end_index: int
                   ) -> list[MutableToken]:
        tkns = []
        for tkn in self:
            if (tkn.base.char_index >= start_index and
                    tkn.base.char_index <= end_index):
                tkns.append(tkn)
        return tkns

    def set_addon_data(self, path: str, val: Any) -> None:
        if not self._delegate.has_extension(path):
            raise UnregisteredDataPathException(self.__class__, path)
        setattr(self._delegate._, path, val)

    def has_addon_data(self, path: str) -> bool:
        return bool(self.get_addon_data(path))

    def get_addon_data(self, path: str) -> Any:
        if not self._delegate.has_extension(path):
            raise UnregisteredDataPathException(self.__class__, path)
        return getattr(self._delegate._, path)

    def get_available_addon_paths(self) -> list[str]:
        return [path for path in self._addon_extension_paths
                if self.has_addon_data(path)]

    @classmethod
    def register_addon_path(cls, path: str, def_val: Any = None,
                            force: bool = True) -> None:
        SpacyDoc.set_extension(path, default=def_val, force=force)
        cls._addon_extension_paths.add(path)

    def __iter__(self) -> Iterator[MutableToken]:
        for tkn in iter(self._delegate):
            yield Token(tkn)

    def isupper(self) -> bool:
        return self._delegate.text.isupper()

    def __str__(self):
        return "M2W[T]:" + str(self._delegate)

    def __repr__(self):
        return repr(str(self))
